(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import List Relations.

Require Import utils cr pos.
Require Import uterm uterm_beta uterm_beta_confluent.

Set Implicit Arguments.

Definition ut_beta_eq n := clos_refl_sym_trans _ (@ut_beta n).

Notation " x '~b' y " := (@ut_beta_eq _ x y) (at level 71).

Fact ut_beta_eq_refl n (f : uterm n) : f ~b f.
Proof. constructor 2. Qed.

Fact ut_eq_beta_eq n (f g : uterm n) : f = g -> f ~b g.
Proof. intros []; constructor 2. Qed.

Definition ut_beta_eq_refl' := ut_eq_beta_eq.

Fact ut_beta_eq_sym n (f g : uterm n) : f ~b g -> g ~b f.
Proof. intros; constructor 3; auto. Qed.

Fact ut_beta_eq_trans n (f g h : uterm n) : f ~b g -> g ~b h -> f ~b h.
Proof. intros; constructor 4 with g; auto. Qed.

Fact ut_beta_rt_eq n (f g : uterm n) : f -b>> g -> f ~b g.
Proof. intros; apply clos_rt_clos_rst; auto. Qed.

Definition ut_beta_beta_eq := ut_beta_rt_eq.

Fact ut_beta_redex_eq n f (a h : uterm n) : f[a] ~b h -> \\f @ a ~b h.
Proof.
  intros H.
  apply ut_beta_eq_trans with (2 := H).
  apply ut_beta_rt_eq.
  constructor 1; apply in_ut_beta_redex.
Qed.

Fact ut_beta_eq_abs n (f g : uterm (S n)) : f ~b g -> \\ f ~b \\ g.
Proof.
   apply clos_refl_sym_trans_map with (f := @ut_abs _).
   intros ? ?; apply in_ut_beta_abs.
Qed.

Section ut_beta_eq_app.

  Let ut_beta_eq_app_lft n (f f' g : uterm n) : f ~b f'-> f @ g ~b f' @ g.
  Proof.
    apply clos_refl_sym_trans_map with (f := fun f => f @ g).
    intros ? ?; apply in_ut_beta_lft.
  Qed.

  Let ut_beta_eq_app_rt n (f g g' : uterm n) : g ~b g'-> f @ g ~b f @ g'.
  Proof.
    apply clos_refl_sym_trans_map with (f := fun g => f @ g).
    intros ? ?; apply in_ut_beta_rt.
  Qed.

  Fact ut_beta_eq_app n (f f' g g' : uterm n) : f ~b f' -> g ~b g' -> f @ g ~b f' @ g'.
  Proof.
    intros.
    apply ut_beta_eq_trans with (f @ g').
    apply ut_beta_eq_app_rt; auto.
    apply ut_beta_eq_app_lft; auto.
  Qed.

End ut_beta_eq_app.

Fact ut_beta_eq_ren n (f g : uterm n) m (r : pos_map n m) : f ~b g -> ut_ren f r ~b ut_ren g r.
Proof.
  induction 1.
  constructor 1; apply ut_beta_ren; auto.
  constructor 2.
  constructor 3; auto.
  constructor 4 with (ut_ren y r); auto.
Qed.

Fact ut_beta_eq_app_rec n l (f g : uterm n) : f ~b g -> f @* l ~b g @* l.
Proof.
  apply clos_refl_sym_trans_map with (f := fun x => x @* l).
  intros ? ?; apply ut_beta_app_rec.
Qed. 

Fact ut_beta_eq_app_rev n l (f g : uterm n) : f ~b g -> f *@ l ~b g *@ l.
Proof.
  do 2 rewrite ut_app_rev_rec_eq.
  apply ut_beta_eq_app_rec.
Qed.

Fact ut_beta_eq_lift_rec n m (f g : uterm n) : f ~b g -> ut_lift_rec m f ~b ut_lift_rec m g.
Proof.
  apply clos_refl_sym_trans_map with (f := ut_lift_rec m).
  intros ? ?; apply ut_beta_lift_rec.
Qed.

Fact ut_beta_eq_app_rec_rev_fall n f l1 l2 : Forall2 (@ut_beta_eq n) l2 l1 -> f *@ l1 ~b f *@ l2.
Proof.
  intros H; revert H f.
  induction 1; intros; simpl.
  constructor 2.
  apply ut_beta_eq_app; auto.
  apply ut_beta_eq_sym; auto.
Qed.

Fact ut_beta_eq_app_rec_fall n f l1 l2 : Forall2 (@ut_beta_eq n) l2 l1 -> f @* l1 ~b f @* l2.
Proof.
  intros H.
  do 2 rewrite ut_app_rec_rev_eq.
  apply ut_beta_eq_app_rec_rev_fall.
  revert H; clear f.
  induction 1; simpl; auto.
  apply Forall2_app; auto.
Qed.

Fact ut_beta_eq_cr n (f g : uterm n) : f ~b g <-> exists h, f -b>> h /\ g -b>> h.
Proof.
  split.

  induction 1 as [ f g H1 | f | g f H1 (h & H2 & H3) | f g h H1 (h1 & ? & ?) H2 (h2 & ? & ?) ].
  exists g; split; [ constructor 1 | constructor 2 ]; auto.
  exists f; split; constructor 2.
  exists h; split; auto.
  destruct (ut_beta_rt_cr H0 H3) as (h3 & ? & ?).
  exists h3; split.
  constructor 3 with h1; auto.
  constructor 3 with h2; auto.
  
  intros (h & H1 & H2).
  constructor 4 with h. 
  apply clos_rt_clos_rst; auto.
  constructor 3.
  apply clos_rt_clos_rst; auto.
Qed.

Fact ut_beta_eq_inv_lift n (f g : uterm n) : ut_lift f ~b ut_lift g -> f ~b g.
Proof.
  intros H.
  apply ut_beta_eq_cr in H.
  destruct H as (h & H1 & H2).
  apply ut_beta_rt_lift_inv in H1.
  apply ut_beta_rt_lift_inv in H2.
  destruct H1 as (h1 & ? & H1); subst.
  destruct H2 as (h2 & E & H2).
  apply ut_lift_inj in E; subst.
  apply ut_beta_eq_cr; exists h2; auto.
Qed.

Fact ut_beta_eq_var_app_rev n (p q : pos n) l m : £ p *@ l ~b £ q *@ m -> p = q /\ Forall2 (@ut_beta_eq _) l m.
Proof.
  intros H.
  apply ut_beta_eq_cr in H.
  destruct H as (v & H1 & H2).
  apply ut_beta_rt_var_app_rev_inv in H1.
  apply ut_beta_rt_var_app_rev_inv in H2.
  destruct H1 as (m1 & H3 & H4).
  destruct H2 as (m2 & H1 & H2).
  subst.
  apply ut_var_app_rev_inj in H1.
  destruct H1; subst.
  split; auto.
  apply Forall2_sym in H4.
  generalize (Forall2_comp H4 H2).
  apply Forall2_incl.
  intros x y (z & Hz1 & Hz2).
  apply ut_beta_eq_trans with z.
  apply ut_beta_rt_eq; auto.
  apply ut_beta_eq_sym, ut_beta_rt_eq; auto.
Qed.

Fact ut_beta_eq_var_app_rec n (p q : pos n) l m : £ p @* l ~b £ q @* m -> p = q /\ Forall2 (@ut_beta_eq _) l m.
Proof.
  intros H.
  do 2 rewrite ut_app_rec_rev_eq in H.
  apply ut_beta_eq_var_app_rev in H.
  destruct H as (H1 & H).
  apply Forall2_rev in H.
  do 2 rewrite rev_involutive in H.
  split; auto.
Qed.
 
    
    



