(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import List Relations Arith Omega.

Require Import utils cr pos finitary_rel.
Require Import uterm uterm_beta uterm_normal uterm_eta.

Set Implicit Arguments.

Definition ut_beta_eta n := (@ut_beta n) cup2 (@ut_eta n).

Infix "-be->" := (@ut_beta_eta _) (at level 71).

Fact in_ut_beta_eta_redex_beta n f (a : uterm n) : \\ f @ a -be-> f[a].
Proof.
  left; apply in_ut_beta_redex.
Qed.

Fact in_ut_beta_eta_redex_eta n (f : uterm n) : \\ (ut_lift f @ £ pos0) -be-> f.
Proof.
  right; apply in_ut_eta_redex.
Qed.

Fact in_ut_beta_eta_lft n (x y k : uterm n) : x -be-> y -> x @ k -be-> y @ k.
Proof.
  intros []; [ left | right ].
  apply in_ut_beta_lft; auto.
  apply in_ut_eta_lft; auto.
Qed.

Fact in_ut_beta_eta_rt n (x y k : uterm n) : x -be-> y -> k @ x -be-> k @ y.
Proof.
  intros []; [ left | right ].
  apply in_ut_beta_rt; auto.
  apply in_ut_eta_rt; auto.
Qed.

Fact in_ut_beta_eta_abs n (x y : uterm (S n)) : x -be-> y -> \\ x -be-> \\ y.
Proof.
  intros []; [ left | right ].
  apply in_ut_beta_abs; auto.
  apply in_ut_eta_abs; auto.
Qed.

Definition ut_beta_eta_rt n := clos_refl_trans _ (@ut_beta_eta n).

Infix "-be>>" := (@ut_beta_eta_rt _) (at level 71).

Fact ut_beta_rt_inc_beta_eta_rt n : @ut_beta_rt n inc2 @ut_beta_eta_rt _.
Proof.
  apply clos_refl_trans_mono; left; auto.
Qed.

Fact ut_eta_rt_inc_beta_eta_rt n : @ut_eta_rt n inc2 @ut_beta_eta_rt _.
Proof.
  apply clos_refl_trans_mono; right; auto.
Qed.

Fact ut_beta_eta_rt_refl n (f : uterm n) : f -be>> f.
Proof. constructor 2. Qed.

Fact ut_eq_beta_eta_rt n (f g : uterm n) : f = g -> f -be>> g.
Proof. intros []; constructor 2. Qed.

Fact ut_beta_eta_rt_trans n (f g h : uterm n) : f -be>> g -> g -be>> h -> f -be>> h.
Proof. intros; constructor 3 with g; auto. Qed.

Fact ut_beta_eta_beta_eta_rt n (f g : uterm n) : f -be-> g -> f -be>> g.
Proof. intros; constructor 1; auto. Qed.

Fact ut_beta_eta_rt_inv n (f g : uterm n) : f -be>> g -> f = g \/ exists h, f -be-> h /\ h -be>> g.
Proof. apply clos_refl_trans_step. Qed.

Fact ut_beta_eta_redex_eta_rt n f (a h : uterm n) : f -be>> h -> \\ (ut_lift f @ £ pos0) -be>> h.
Proof.
  intros H.
  apply ut_beta_eta_rt_trans with (2 := H).
  constructor 1; apply in_ut_beta_eta_redex_eta.
Qed.

Fact ut_beta_eta_redex_beta_rt n f (a h : uterm n) : f[a] -be>> h -> \\f @ a -be>> h.
Proof.
  intros H.
  apply ut_beta_eta_rt_trans with (2 := H).
  constructor 1; apply in_ut_beta_eta_redex_beta.
Qed.

Fact ut_beta_eta_rt_lft n (f g a : uterm n) : f -be>> g -> f @ a -be>> g @ a.
Proof.
  apply clos_refl_trans_map with (f := fun x => x @ a).
  intros ? ?; apply in_ut_beta_eta_lft.
Qed.

Fact ut_beta_eta_rt_rt n (f a b : uterm n) : a -be>> b -> f @ a -be>> f @ b.
Proof.
  apply clos_refl_trans_map with (f := fun x => f @ x).
  intros ? ?; apply in_ut_beta_eta_rt.
Qed.

Fact ut_beta_eta_rt_app n (f g a b : uterm n) : f -be>> g -> a -be>> b -> f @ a -be>> g @ b.
Proof.
  intros H1 H2.
  constructor 3 with (f @ b).
  apply ut_beta_eta_rt_rt; auto.
  apply ut_beta_eta_rt_lft; auto.
Qed.

Fact ut_beta_eta_rt_abs n (f g : uterm (S n)) : f -be>> g -> \\ f -be>> \\ g.
Proof.
  apply clos_refl_trans_map with (f := @ut_abs _).
  intros ? ?; apply in_ut_beta_eta_abs.
Qed.

Fact ut_beta_eta_subst n (f g : uterm n) m (s : uts_t n m) : f -be-> g -> ut_subst f s -be-> ut_subst g s.
Proof.
  intros []; [ left | right ].
  apply ut_beta_subst; auto.
  apply ut_eta_subst; auto.
Qed.

Fact ut_beta_eta_rt_subst n (f g : uterm n) m (s : uts_t n m) : f -be>> g -> ut_subst f s -be>> ut_subst g s.
Proof.
  apply clos_refl_trans_map with (f := fun x => ut_subst x s).
  intros ? ?; apply ut_beta_eta_subst.
Qed.

Fact ut_beta_eta_reduce n (f g : uterm (S n)) a : f -be-> g -> f[a] -be-> g[a].
Proof.
  apply ut_beta_eta_subst.
Qed.

Fact ut_beta_eta_ren n (f g : uterm n) m (r : pos_map n m) : f -be-> g -> ut_ren f r -be-> ut_ren g r.
Proof. do 2 rewrite ut_ren_subst_eq; apply ut_beta_eta_subst. Qed.

Fact ut_beta_eta_rt_ren n (f g : uterm n) m (r : pos_map n m) : f -be>> g -> ut_ren f r -be>> ut_ren g r.
Proof. do 2 rewrite ut_ren_subst_eq; apply ut_beta_eta_rt_subst. Qed.

Fact ut_beta_eta_lift n (f g : uterm n) : f -be-> g -> ut_lift f -be-> ut_lift g.
Proof. apply ut_beta_eta_ren. Qed.

Fact ut_beta_eta_rt_lift n (f g : uterm n) : f -be>> g -> ut_lift f -be>> ut_lift g.
Proof. apply ut_beta_eta_rt_ren. Qed.

Fact ut_subst_beta_eta_rt n (f : uterm n)  m (r s : uts_t n m) :
  (forall p, r p -be>> s p) -> ut_subst f r -be>> ut_subst f s.
Proof.
  revert m r s.
  induction f as [ n f Hf | n a Ha b Hb | n v ]; intros m r s Hrs; simpl; auto.
  apply ut_beta_eta_rt_abs, Hf.
  intros p; pos_inv p; unfold uts_lift; simpl.
  constructor 2.
  apply ut_beta_eta_rt_lift; auto.
  apply ut_beta_eta_rt_app; auto.
Qed.

Fact ut_reduce_beta_eta_rt n (f : uterm (S n)) a b : a -be>> b -> f[a] -be>> f[b].
Proof.
  intros H; apply ut_subst_beta_eta_rt.
  intros p; unfold uts_first; pos_inv p; auto.
  constructor 2.
Qed.

Fact ut_beta_eta_app_rec n l : forall (f g : uterm n), f -be-> g -> f @* l -be-> g @* l.
Proof.
  induction l as [ | x l IHl ]; simpl; auto; intros f g Hfg.
  apply IHl, in_ut_beta_eta_lft; auto.
Qed.

Fact ut_beta_eta_rt_app_rec n l (f g : uterm n) : f -be>> g -> f @* l -be>> g @* l.
Proof.
  apply clos_refl_trans_map with (f := fun x => x @* l).
  intros ? ?; apply ut_beta_eta_app_rec.
Qed.

Fact ut_beta_eta_app_rec_In n l r : forall (f a b : uterm n), a -be-> b -> f @* (l++a::r) -be-> f @* (l++b::r).
Proof.
  induction l as [ | x l IHl ]; simpl; auto; intros f a b H.
  apply ut_beta_eta_app_rec, in_ut_beta_eta_rt; auto.
Qed.

Fact ut_beta_eta_app_rev n l (f g : uterm n) : f -be-> g -> f *@ l -be-> g *@ l.
Proof.
  do 2 rewrite ut_app_rev_rec_eq.
  apply ut_beta_eta_app_rec.
Qed.

Fact ut_beta_eta_rt_app_rev n l (f g : uterm n) : f -be>> g -> f *@ l -be>> g *@ l.
Proof.
  do 2 rewrite ut_app_rev_rec_eq.
  apply ut_beta_eta_rt_app_rec.
Qed.

Section ut_beta_eta_inv.

  Fact ut_beta_eta_var_inv n (p : pos n) h : ~ £ p -be-> h.
  Proof.
    intros [|].
    apply ut_beta_var_inv.
    apply ut_eta_var_inv.
  Qed.
  
  Fact ut_beta_eta_abs_inv n u (v : uterm n) : 
      \\ u -be-> v ->  u = ut_lift v @ £ pos0 
                   \/ (exists w, v = \\ w /\ u -be-> w).
  Proof.
    intros [ H | H ].
    apply ut_beta_abs_inv in H.
    destruct H as (w & H1 & H2); right; exists w; split; auto; left; auto.
    apply ut_eta_abs_inv in H.
    destruct H as [ H | (w & H1 & H2) ]; [ left | right ]; auto.
    exists w; split; auto; right; auto.
  Qed.
  
  Fact ut_beta_eta_app_inv n (h u v : uterm n) : 
           h @ u -be-> v 
       -> (exists f, h = \\ f /\ v = f[u])
       \/ (exists g, v = g @ u /\ h -be-> g)
       \/ (exists w, v = h @ w /\ u -be-> w).
  Proof.
    intros [ H | H ].
    apply ut_beta_app_inv in H.
    destruct H as [ H | [ (w & H1 & H2) | (w & H1 & H2) ] ]; [ left | right; left | right; right ]; 
      auto; exists w; split; auto; left; auto.
    apply ut_eta_app_inv in H.
    right.
    destruct H as [ (w & H1 & H2) | (w & H1 & H2) ]; [ left | right ];
      exists w; split; auto; right; auto.
  Qed.

  Fact ut_beta_eta_redex_inv n (f : uterm (S n)) (a : uterm n) v :
       \\ f @ a -be-> v ->  v = f[a]
                        \/ (exists g, v = g @ a /\ f = ut_lift g @ £ pos0)
                        \/ (exists g, v = \\ g @ a /\ f -be-> g)
                        \/ (exists b, v = \\ f @ b /\ a -be-> b).
  Proof.
    intros H.
    apply ut_beta_eta_app_inv in H.
    destruct H as [ (g & H & ?) | [ (g & ? & H) | (b & ? & H) ] ]; subst.
    injc H; left; auto.
    apply ut_beta_eta_abs_inv in H.
    destruct H as [ H | (f' & ? & H) ]; subst.
    right; left; subst; exists g; auto.
    do 2 right; left; exists f'; auto.
    do 3 right; exists b; auto.
  Qed.

  Fact ut_beta_eta_neutral_app_inv n (h u v : uterm n) :
           ut_neutral h  
       ->  h @ u -be-> v 
       -> (exists g, v = g @ u /\ h -be-> g)
       \/ (exists w, v = h @ w /\ u -be-> w).
  Proof.
    intros H1 H2.
    apply ut_beta_eta_app_inv in H2.
    destruct H2 as [ (g & H & ?) | [ (g & ? & H) | (b & ? & H) ] ]; subst.
    destruct H1.
    left; exists g; auto.
    right; exists b; auto.
  Qed.

  Fact ut_beta_eta_neutral_app_rev_inv n a (v : uterm n) ll : 
        ut_neutral a
     -> a *@ ll -be-> v -> (exists a', v = a' *@ ll /\ a -be-> a')
                        \/  exists l c c' r, ll = l++c::r /\ v = a *@ (l++c'::r) /\ c -be-> c'.
  Proof.
    revert a v; induction ll as [ | b ll IHll ]; simpl; intros a v H1 H2.
    left; exists v; auto.

    apply ut_beta_eta_neutral_app_inv in H2.
    destruct H2 as [ (f & H2 & H3)
                   | (f & H2 & H3) ]; subst.

    destruct (IHll _ _ H1 H3) as [ (a' & H4 & H5) | (l & c & c' & r & H4 & H5 & H6) ]; subst.
    left; exists a'; auto.
    right; exists (b::l), c, c', r; auto.
    
    right; exists nil, b, f, ll; simpl; auto.

    destruct ll; simpl; auto.
  Qed.

  Fact ut_beta_eta_neutral_app_rec_inv n a b (v : uterm n) ll : 
       (a @ b) @* ll -be-> v -> (exists c,        v = c @* ll 
                                              /\ a @ b  -be-> c)
                             \/  exists l c c' r, ll = l++c::r 
                                              /\ v = (a @ b) @* (l++c'::r) 
                                              /\ c -be-> c'.
  Proof.
    intros H.
    rewrite ut_app_rec_rev_eq in H.
    apply ut_beta_eta_neutral_app_rev_inv in H.
    destruct H as [ (c & ? & H) | (l & c & c' & r & H1 & H2 & H3) ]; subst.
 
    left; exists c; rewrite ut_app_rec_rev_eq; auto.
    right. exists (rev r), c, c', (rev l); repeat split; auto.
    rewrite <- (rev_involutive ll), H1, rev_app_distr; simpl.
    rewrite app_ass; reflexivity.
    rewrite ut_app_rec_rev_eq, rev_app_distr; simpl.
    do 2 rewrite rev_involutive; rewrite app_ass; auto.
    red; auto.
  Qed.
  
End ut_beta_eta_inv.

Section ut_eta_beta_eq.

  Reserved Notation " x '-eb->' y " (at level 71).

  Inductive ut_eta_beta : forall n, uterm n -> uterm n -> Prop :=
    | in_ut_eb_eta   : forall n (f : uterm n) ,  \\ (ut_lift f @ £ pos0) -eb-> f  
    | in_ut_eb_beta  : forall n f (a   : uterm n),              \\ f @ a -eb-> f[a]  
    | in_ut_eb_lft   : forall n (x y k : uterm n),    x -eb-> y -> x @ k -eb-> y @ k
    | in_ut_eb_rt    : forall n (x y k : uterm n),    x -eb-> y -> k @ x -eb-> k @ y
    | in_ut_eb_abs   : forall n (x y : uterm (S n)),  x -eb-> y -> \\ x  -eb-> \\ y 
  where  " x '-eb->' y " := (@ut_eta_beta _ x y).
  
  Theorem ut_eta_beta_eq n (u v : uterm n) : u -be-> v <-> u -eb-> v.
  Proof.
    split.
    intros [].
    
    induction 1.
    constructor 2.
    constructor 3; auto.
    constructor 4; auto.
    constructor 5; auto.
    
    induction 1.
    constructor 1.
    constructor 3; auto.
    constructor 4; auto.
    constructor 5; auto.
    
    induction 1 as [ | | ? ? ? ? ? [] | ? ? ? ? ? [] | ? ? ? ? [] ].
    right; constructor 1.
    left; constructor 1.
    left; constructor 2; auto.
    right; constructor 2; auto.
    left; constructor 3; auto.
    right; constructor 3; auto.
    left; constructor 4; auto.
    right; constructor 4; auto.
  Qed.
  
End ut_eta_beta_eq.

Section ut_beta_eta_normal.

  Fact ut_neutral_eta n (f g : uterm n) : f -e-> g -> ut_neutral f -> ut_neutral g.
  Proof.
    induction 1; simpl; tauto.
  Qed.

  Fact ut_normal_eta n (f g : uterm n) : f -e-> g -> ut_normal f -> ut_normal g.
  Proof.
    induction 1 as [ | ? ? ? ? H | ? ? ? ? H | ]; intros G.

    apply ut_abs_normal, ut_normal_app, proj2, proj1 in G.
    apply ut_normal_lift; auto.

    apply ut_normal_app in G.
    apply ut_app_normal; try tauto.
    apply ut_neutral_eta with (1 := H); tauto.

    apply ut_normal_app in G.
    apply ut_app_normal; tauto.
  
    apply ut_abs_normal.
    apply ut_abs_normal in G; auto.
  Qed.

  Fact ut_normal_eta_rt n (f g : uterm n) : f -e>> g -> ut_normal f -> ut_normal g.
  Proof.
    induction 1; auto; apply ut_normal_eta; auto.
  Qed.

  Definition ut_beta_eta_normal n (f : uterm n) := forall g, ~ f -be-> g.

  Fact ut_beta_eta_normal_form n f : @ut_normal n f -> { g | f -be>> g /\ ut_beta_eta_normal g }.
  Proof.
    intros H.
    destruct (ut_eta_normal_form f) as (g & H1 & H2).
    exists g; split.
    apply ut_eta_rt_inc_beta_eta_rt; auto.
    generalize (ut_normal_eta_rt H1 H); intros H3.
    intros h [ | ].
    apply H3.
    apply H2.
  Qed.

  Fact ut_beta_eta_normal_beta_normal n : @ut_beta_eta_normal n inc1 @ut_beta_normal _.
  Proof.
    intros f Hf g Hg; apply (Hf g); left; auto.
  Qed.

  Definition ut_beta_eta_normalizable n (f : uterm n) := exists g, f -be>> g /\ ut_beta_eta_normal g.

  Fact ut_normalizable_beta_eta_normalizable n f :
      @ut_normalizable n f -> ut_beta_eta_normalizable f.
  Proof.
    intros (g & H1 & H2).
    destruct (ut_beta_eta_normal_form H2) as (h & H3 & H4).
    exists h; split; auto.
    apply ut_beta_eta_rt_trans with g; auto.
    apply ut_beta_rt_inc_beta_eta_rt; auto.
  Qed.

  Fact ut_eta_app_left_rel_iter n (u v a : uterm n) i :
      rel_iter (@ut_eta n) i u v -> rel_iter (@ut_eta n) i (u @ a) (v @ a).
  Proof.
    revert u v.
    induction i as [ | i IHi ]; intros u v; simpl.
    intros []; auto. 
    intros (k & H1 & H2).
    apply IHi in H2.
    exists (k @ a); split; auto.
    constructor 2; auto.
  Qed.

  Fact ut_beta_app_left_rel_iter n (u v a : uterm n) i :
      rel_iter (@ut_beta n) i u v -> rel_iter (@ut_beta n) i (u @ a) (v @ a).
  Proof.
    revert u v.
    induction i as [ | i IHi ]; intros u v; simpl.
    intros []; auto. 
    intros (k & H1 & H2).
    apply IHi in H2.
    exists (k @ a); split; auto.
    constructor 2; auto.
  Qed.

  Fact ut_beta_app_right_rel_iter n (u v a : uterm n) i :
      rel_iter (@ut_beta n) i u v -> rel_iter (@ut_beta n) i (a @ u) (a @ v).
  Proof.
    revert u v.
    induction i as [ | i IHi ]; intros u v; simpl.
    intros []; auto. 
    intros (k & H1 & H2).
    apply IHi in H2.
    exists (a @ k); split; auto.
    constructor 3; auto.
  Qed.

  Fact ut_beta_abs_rel_iter n (f g : uterm (S n)) i :
      rel_iter (@ut_beta _) i f g -> rel_iter (@ut_beta _) i (\\ f) (\\ g).
  Proof.
    revert f g.
    induction i as [ | i IHi ]; intros f g; simpl.
    intros []; auto. 
    intros (k & H1 & H2).
    apply IHi in H2.
    exists (\\ k); split; auto.
    constructor 4; auto.
  Qed.
  
  (* This is the best we can do ... I am not sure it is possible to show
     beta-SN <-> beta-eta-SN that way *)

  Fact ut_beta_inv_eta_square n (f g h : uterm n) : 
    f -e-> g -> g -b-> h -> exists u i, 1 <= i <= 2 /\ rel_iter (@ut_beta n) i f u /\ u -e>> h. 
  Proof.
    intros H; revert H h.
    induction 1 as [ n f a | n f h a H1 IH | n f a b H1 IH | n f h H IH ]; intros v Hv.

    exists (\\ (ut_lift v @ £ pos0)), 1; repeat split; auto.
    simpl.
    exists (\\ (ut_lift v @ £ pos0)); split; auto.
    constructor 4; constructor 2.
    apply ut_beta_lift; auto.
    constructor 1; constructor 1.

    apply ut_beta_app_inv in Hv.
    destruct Hv as [ (g & ? & ?) | [ (g & ? & Hg) | (g & ? & Hg) ] ]; subst.

    apply ut_rev_eta_abs_inv in H1.
    destruct H1 as [ (f' & ? & H1) | H1 ]; subst.
 
    exists (f'[a]), 1; repeat split; auto.
    exists (f'[a]); split; auto; constructor.
    apply ut_eta_eta_rt, ut_eta_subst; auto.

    (* here we need two beta reductions *)

    exists (g[a]), 2.
    repeat split; auto.
    exists (\\ g @ a); split.
    replace (\\ g @ a) with ( (ut_lift (\\g) @ £ pos0)[a] ).
    constructor 1.
    rewrite ut_reduce_app; f_equal; apply ut_reduce_lift.
    exists (g[a]); split; simpl; auto.
    constructor 1.
    apply ut_eta_rt_refl.

    destruct (IH _ Hg) as (u & i & (? & ?) & ? & ?).
    exists (u @ a), i; repeat split; auto.
    apply ut_beta_app_left_rel_iter; auto.
    apply ut_eta_rt_lft; auto.

    exists (f @ g), 1; repeat split; auto.
    exists (f @ g); simpl; repeat split; auto.
    constructor 3; auto.
    apply ut_eta_rt_lft; constructor 1; auto.
    
    apply ut_beta_app_inv in Hv.
    destruct Hv as [ (w & ? & Hw) | [ (w & ? & Hw) | (w & ? & Hw) ] ]; subst.
    
    (* here we need a large amount of eta reductions due
       to the potential substitution of an eta redex *)

    exists (w[f]), 1.
    repeat split; auto.
    exists (w[f]); simpl; split; auto.
    constructor 1.
    apply ut_eta_rt_reduce; constructor 1; auto.
    
    exists (w @ f), 1; repeat split; auto.
    exists (w @ f); simpl; split; auto.
    constructor 2; auto.
    apply ut_eta_rt_rt; constructor 1; auto.

    destruct (IH _ Hw) as (u & i & (? & ?) & ? & ?).
    exists (b @ u), i; repeat split; auto.
    apply ut_beta_app_right_rel_iter; auto.
    apply ut_eta_rt_rt; auto.
    
    apply ut_beta_abs_inv in Hv.
    destruct Hv as (g & ? & Hg); subst.
    destruct (IH _ Hg) as (u & i & (? & ?) & ? & ?).
    exists (\\ u), i; repeat split; auto.
    apply ut_beta_abs_rel_iter; auto.
    apply ut_eta_rt_abs; auto.
  Qed.

End ut_beta_eta_normal.

Fact ut_beta_eta_ren_rel_comm n m (R : ren_rel n m) : 
       square (@ut_beta_eta _) (ut_ren_rel R) (ut_ren_rel R) (@ut_beta_eta _).
Proof.
  intros f g u [ H1 | H1 ] H2.
  destruct (ut_beta_ren_rel_comm H1 H2) as (h & ? & ?). 
  exists h; split; auto; left; auto.
  destruct (ut_eta_ren_rel_comm H1 H2) as (h & ? & ?). 
  exists h; split; auto; right; auto.
Qed.

Section implementation_of_SN.

  Variable (n : nat).
  
  Implicit Type (u : uterm n).

  (* Hence the direct image of beta-eta reduction is computed into a list *)

  Definition ut_beta_eta_list u := proj1_sig (@ut_beta_finite n u) ++ proj1_sig (@ut_eta_finite n u).

  Definition ut_beta_eta_list_prop u : forall v, u -be-> v <-> In v (ut_beta_eta_list u).
  Proof.
    unfold ut_beta_eta_list.
    intros v; split.
    intros [ H | H ]; apply in_or_app; [ left | right ]; revert H.
    apply (proj2_sig (@ut_beta_finite n u)).
    apply (proj2_sig (@ut_eta_finite n u)).
    intros H; apply in_app_or in H; 
      destruct H as [ H | H ]; [ left | right ]; revert H.
    apply (proj2_sig (@ut_beta_finite n u)).
    apply (proj2_sig (@ut_eta_finite n u)).
  Qed.

  (* Strongly normalising terms are those which are accessible under the
     inverse of beta-eta reduction *) 

  Definition ut_beta_eta_sn := Acc (inv (@ut_beta_eta n)).

  Notation SN := ut_beta_eta_sn.

  Fact ut_beta_eta_beta_eta_sn f g : 
       f -be-> g -> ut_beta_eta_sn f -> ut_beta_eta_sn g.  
  Proof.
    intros H1 H2; apply (Acc_inv H2 _ H1).
  Qed.

  Fact ut_beta_eta_beta_eta_rt_sn f g : 
       f -be>> g -> ut_beta_eta_sn f -> ut_beta_eta_sn g.
  Proof.
    induction 1; auto; apply ut_beta_eta_beta_eta_sn; auto.
  Qed.

  (* Property 1 is for free *)

  Fact ut_beta_eta_sn_fix u : (forall v, u -be-> v -> SN v) <-> SN u.
  Proof.
    split.
    apply Acc_intro with (R := inv (@ut_beta_eta _)).
    apply Acc_inv. 
  Qed.

  (* there is a path u -b-> ... -b-> v of length p *)

  Notation "u '-[' p ']->' v" := (rel_iter (fun x y => x -be-> y) p u v) (at level 70).

  (* The height (maximum length of a reduction) can be computed from the
     SN/accessibility property, see finitary_rel.v for the generic
     implementation *)

  Definition ut_beta_eta_sn_ht u : SN u -> nat.
  Proof. 
    apply fin_rel_ht with (f := ut_beta_eta_list).
    intros; apply ut_beta_eta_list_prop; tauto.
  Defined.

  (* the height needs a proof of SN/accessibility but the result does not depend on it *)

  Fact ut_beta_eta_sn_ht_eq u (H1 H2 : SN u) : ut_beta_eta_sn_ht H1 = ut_beta_eta_sn_ht H2.
  Proof. apply fin_rel_ht_eq. Qed.

  (* the height is an upper bound *)

  Fact ut_beta_eta_sn_ht_bound u (Hu : SN u) p v : u -[p]-> v -> p <= ut_beta_eta_sn_ht Hu.
  Proof. apply fin_rel_ht_bound. Qed.

  (* and this upper bound is reached for some computable term *)

  Fact ut_beta_eta_sn_ht_lub u (Hu : SN u) : { v | u -[ut_beta_eta_sn_ht Hu]-> v }.
  Proof. apply fin_rel_ht_lub. Qed.

  (* We deduce Property 2 *)

  Fact ut_beta_eta_sn_dec u (Hu : SN u) v (Hv : SN v) : 
      u -be-> v -> ut_beta_eta_sn_ht Hv < ut_beta_eta_sn_ht Hu.
  Proof.
    intros H.
    destruct (ut_beta_eta_sn_ht_lub Hv) as (w & Hw).
    apply ut_beta_eta_sn_ht_bound with w.
    exists v; auto.
  Qed.

  Fact ut_beta_eta_rt_sn_dec u (Hu : SN u) v (Hv : SN v) : 
      u -be>> v -> ut_beta_eta_sn_ht Hv <= ut_beta_eta_sn_ht Hu.
  Proof.
    intros H; revert H Hu Hv.
    induction 1 as [ | | u v w H1 IH1 H2 IH2 ]; intros Hu Hv.
    apply lt_le_weak, ut_beta_eta_sn_dec; auto.
    rewrite (ut_beta_eta_sn_ht_eq Hu Hv); auto.
    apply le_trans with (2 := IH1 Hu (ut_beta_eta_beta_eta_rt_sn H1 Hu)); auto.
  Qed.

End implementation_of_SN.

Implicit Arguments ut_beta_eta_sn_ht [ n ].

Fact ut_beta_eta_ren_sn n m r f : ut_beta_eta_sn f <-> ut_beta_eta_sn (@ut_ren n f m r).
Proof.
  split.

  apply square_Acc with (E := ut_ren_rel (fun p q => p = r q)).
  apply ut_beta_eta_ren_rel_comm.
  apply ut_ren_rel_sym, ut_ren_rel_eq_ren.

  apply square_Acc with (E := ut_ren_rel (fun p q => q = r p)).
  apply ut_beta_eta_ren_rel_comm.
  apply ut_ren_rel_eq_ren.
Qed.

Corollary ut_beta_eta_lift_sn n f : ut_beta_eta_sn f <-> ut_beta_eta_sn (@ut_lift n f).
Proof.
  apply ut_beta_eta_ren_sn.
Qed.

Fact ut_beta_eta_sn_app_neutral n (f a : uterm n) : 
        (forall f', f -be>> f' -> ut_neutral f') 
      -> ut_beta_eta_sn f 
      -> ut_beta_eta_sn a
      -> ut_beta_eta_sn (f @ a).
Proof.
  intros H1 H2 H3.
  revert H2 a H3 H1.
  induction 1 as [ f Hf IHf ].
  induction 1 as [ a Ha IHa ].
  intros H.
  constructor 1; intros v Hv.
  apply ut_beta_eta_neutral_app_inv in Hv.
  2: apply H; constructor 2.
  destruct Hv as [ (g & ? & Hg) | (b & ? & Hb) ]; subst.
  apply (IHf _ Hg); auto.
  constructor; apply Ha.
  intros h Hh; apply H.
  constructor 3 with g; auto; constructor 1; auto.
  apply (IHa _ Hb); auto.
Qed.

Fact ut_beta_eta_neutral_normal_app_rev n (f : uterm n) ll g : 
     ut_neutral f 
  -> ut_beta_eta_normal f 
  -> f *@ ll -be-> g 
  -> exists l a b r, ll = l++a::r /\ a -be-> b /\ g = f *@ l++b::r.
Proof.
  intros H1 H2 H3.
  
  apply ut_beta_eta_neutral_app_rev_inv in H3; auto.
  destruct H3 as [ (f' & _ & H3) | (l & a & b & r & ? & ? & H3) ]; subst.
  apply H2 in H3; destruct H3.
  exists l, a, b, r; auto.
Qed.

Fact ut_beta_eta_rt_neutral_normal_app_rev n (f : uterm n) ll g : 
     ut_neutral f 
  -> ut_beta_eta_normal f 
  -> f *@ ll -be>> g 
  -> exists mm,  Forall2 (@ut_beta_eta_rt _) ll mm /\ g = f *@ mm.
Proof.
  intros H1 H2 H.
  apply clos_rt_rtn1 in H.
  induction H as [ | g h H3 H4 IH4 ].
  exists ll; split; auto.
  induction ll; constructor; auto; constructor 2.
  destruct IH4 as (mm & H5 & H6); subst.
  apply ut_beta_eta_neutral_normal_app_rev in H3; auto.
  destruct H3 as (l & a & b & r & H3 & H6 & H7); subst.
  exists (l++b::r); split; auto.
  clear H4.
  assert (Forall2 (@ut_beta_eta_rt _) (l++a::r) (l++b::r)) as H4.
    clear H5.
    apply Forall2_app.
    induction l; constructor; auto; constructor 2.
    constructor.
    constructor 1; auto.
    induction r; constructor; auto; constructor 2.
  generalize (Forall2_comp H5 H4).
  apply Forall2_incl.
  intros ? ? (z & ? & ?).
  apply ut_beta_eta_rt_trans with z; auto.
Qed.
 
Fact ut_beta_eta_rt_app_rev_neutral n (f : uterm n) l g : 
  ut_neutral f -> ut_beta_eta_normal f -> f *@ l -be>> g -> ut_neutral g.
Proof.
  intros Hf1 Hf2 H.
  apply ut_beta_eta_rt_neutral_normal_app_rev in H; auto.
  destruct H as (mm & _ & ?); subst.
  destruct mm; simpl; auto.
Qed.

Fact ut_beta_eta_normal_sn n (f : uterm n) : ut_beta_eta_normal f -> ut_beta_eta_sn f.
Proof.
  intros H; constructor 1; intros g Hg.
  apply H in Hg; destruct Hg.
Qed.

Fact ut_beta_eta_sn_app_rev n (f : uterm n) l :
     ut_neutral f
  -> ut_beta_eta_normal f
  -> Forall (@ut_beta_eta_sn _) l -> ut_beta_eta_sn (f *@ l).
Proof.
  intros Hf1 Hf2.
  induction 1 as [ | a l H1 H2 IH2 ]; simpl.
  apply ut_beta_eta_normal_sn; auto.
  apply ut_beta_eta_sn_app_neutral; auto.
  intro; apply ut_beta_eta_rt_app_rev_neutral; auto.
Qed.

Fact ut_beta_eta_sn_app_rec n (f : uterm n) l :
     ut_neutral f
  -> ut_beta_eta_normal f
  -> Forall (@ut_beta_eta_sn _) l -> ut_beta_eta_sn (f @* l).
Proof.
  rewrite Forall_rev, ut_app_rec_rev_eq.
  apply ut_beta_eta_sn_app_rev.
Qed.

Fact ut_var_beta_eta_normal n (v : pos n) : ut_beta_eta_normal (£ v).
Proof.
  intros ? H; apply ut_beta_eta_var_inv in H; auto.
Qed.

Fact ut_beta_eta_sn_var_app_rev n (p : pos n) l :
     Forall (@ut_beta_eta_sn _) l -> ut_beta_eta_sn (£ p *@ l).
Proof.
  apply ut_beta_eta_sn_app_rev.
  red; auto.
  apply ut_var_beta_eta_normal.
Qed.

Fact ut_beta_eta_sn_left_inv n (a b : uterm n) : ut_beta_eta_sn (a @ b) -> ut_beta_eta_sn a.
Proof.
  gen_eq (a @ b).
  intros f E H; revert f H a b E.
  induction 1 as [ f Hf IHf ]; intros a b ?; subst.
  constructor 1; intros a' Ha'.
  apply IHf with (b := b) (2 := eq_refl).
  apply in_ut_beta_eta_lft; auto.
Qed.

Fact ut_beta_sn_right_inv n (a b : uterm n) : ut_beta_eta_sn (a @ b) -> ut_beta_eta_sn b.
Proof.
  gen_eq (a @ b).
  intros f E H; revert f H a b E.
  induction 1 as [ f Hf IHf ]; intros a b ?; subst.
  constructor 1; intros b' Hb'.
  apply IHf with (a := a) (2 := eq_refl).
  apply in_ut_beta_eta_rt; auto.
Qed.

Section ut_beta_eta_sn_redex.

  Variable (n : nat).

  Let P p q := forall (a : uterm n) u l Ha Ht,
            (ut_beta_eta_sn_ht    a       Ha <= p) 
         -> (ut_beta_eta_sn_ht (u[a] @*l) Ht <= q)
         -> ut_beta_eta_sn (\\ u @* a :: l).

  Let ut_beta_eta_sn_redex_rec p q : P p q.
  Proof.
    revert p q; apply ut_sn_lexico; unfold P.
    intros p q Hpq a u ll Ha Ht H1 H2.
    constructor 1; intros v Hv.
    simpl in Hv.
    apply ut_beta_eta_neutral_app_rec_inv in Hv.
    destruct Hv as [ (w & ? & Hb) | (l & b & b' & r & G1 & G2 & G3) ]; subst.
    apply ut_beta_eta_redex_inv in Hb.
    destruct Hb as [ ? 
                 | [ (g & ? & ?)
                 | [ (g & ? & Hb) 
                   | (b & ? & Hb) ] ] ]; subst.

    apply Ht.
   
    generalize Ht; eqgoal; unfold ut_beta_eta_sn. 
    do 2 f_equal.
    unfold ut_reduce; simpl; f_equal.
    rewrite ut_subst_lift_first; auto.
   
    apply ut_beta_eta_reduce with (a := a) in Hb.
    apply ut_beta_eta_app_rec with (l := ll) in Hb.
    generalize (Acc_inv Ht _ Hb); intros Ht'.
    generalize (ut_beta_eta_sn_dec Ht Ht' Hb); intros H3.
    specialize (Hpq p (ut_beta_eta_sn_ht _ Ht')).
    apply Hpq with (Ha := Ha) (Ht := Ht'); auto.
    right; split; auto; omega.
      
    generalize (Acc_inv Ha _ Hb); intros Ha'.
    generalize (ut_beta_eta_sn_dec Ha Ha' Hb); intros H3.
    generalize Hb; intros Ht'.
    apply ut_beta_eta_beta_eta_rt in Ht'.
    apply ut_reduce_beta_eta_rt with (f := u) in Ht'.
    apply ut_beta_eta_rt_app_rec with (l := ll) in Ht'.
    generalize (ut_beta_eta_beta_eta_rt_sn Ht' Ht); intros Ht''.
    specialize (Hpq (ut_beta_eta_sn_ht _ Ha') q).
    apply Hpq with (Ha := Ha') (Ht := Ht''); auto.
    left; omega.
    apply le_trans with (2 := H2).
    apply ut_beta_eta_rt_sn_dec; auto.

    generalize (ut_beta_eta_app_rec_In l r (u[a]) G3); intros H3.
    generalize (Acc_inv Ht _ H3); intros Ht'.
    generalize (ut_beta_eta_sn_dec Ht Ht' H3); intros H4.
    specialize (Hpq p (ut_beta_eta_sn_ht _ Ht')).
    apply Hpq with (Ha := Ha) (Ht := Ht'); auto.
    right; split; auto; omega.
  Qed.      

  Fact ut_beta_eta_sn_redex (u : uterm (S n)) a l : 
               ut_beta_eta_sn a 
            -> ut_beta_eta_sn ((u[a]) @* l) 
            -> ut_beta_eta_sn (\\ u @* a :: l).
  Proof.
    intros Ha Ht.
    apply (@ut_beta_eta_sn_redex_rec (ut_beta_eta_sn_ht _ Ha) 
                                     (ut_beta_eta_sn_ht _ Ht) 
                                      _ _ _ Ha Ht); auto.
  Qed.
  
End ut_beta_eta_sn_redex.




