(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import Arith List Relations.

Require Import utils cr pos.
Require Import uterm
               uterm_beta uterm_beta_confluent 
               uterm_eta  uterm_eta_confluent
               uterm_beta_eta.

Set Implicit Arguments.

Local Theorem ut_beta_eta_square n (f : uterm n) a b :
       f -b-> a
    -> f -e-> b
    -> exists g, a -e>> g /\ (b -b-> g \/ b = g).
Proof.
  intros H; revert H b.
  induction 1 as [ n f a | n a b c H IH | n a b c H IH | n f g H IH ].

  (* \\ f @ a -b-> f[a] *)

  intros b Hb.
  apply ut_eta_app_inv in Hb.
  destruct Hb as [ (g & ? & Hg) | (w & ? & Hw) ]; subst b.
  apply ut_eta_abs_inv in Hg.
  destruct Hg as [ ? | (w & ? & Hw) ]; subst.

  exists (g @ a); split.
  rewrite ut_reduce_app, ut_reduce_pos0, ut_reduce_lift.
  apply ut_eta_rt_refl.
  right; auto.

  exists (w[a]).
  split.
  apply ut_eta_rt_subst, ut_eta_eta_rt; auto.
  left; constructor.
  
  exists (f[w]).
  split.
  apply ut_eta_rt_reduce, ut_eta_eta_rt; auto.
  left; constructor.

  (* a @ c -b-> b @ c *)

  intros d Hd.
  apply ut_eta_app_inv in Hd.
  destruct Hd as [ (g & ? & Hg) | (w & ? & Hw) ]; subst d.
  
  destruct (IH _ Hg) as (k & H1 & H2).
  exists (k @ c); split.
  apply ut_eta_rt_lft; auto.
  destruct H2 as [ H2 | H2 ].
  left; constructor 2; auto.
  subst; right; auto.
  
  exists (b @ w); split.
  apply ut_eta_rt_rt, ut_eta_eta_rt; auto.
  left; constructor 2; auto.

  (* c @ a -b-> c @ b *)

  intros d Hd.
  apply ut_eta_app_inv in Hd.
  destruct Hd as [ (g & ? & Hg) | (w & ? & Hw) ]; subst d.
  
  exists (g @ b); split.
  apply ut_eta_rt_lft, ut_eta_eta_rt; auto.
  left; constructor 3; auto.

  destruct (IH _ Hw) as (k & H1 & H2).
  exists (c @ k); split.
  apply ut_eta_rt_rt; auto.
  destruct H2 as [ H2 | H2 ].
  left; constructor 3; auto.
  subst; right; auto.

  (* \\ f -b-> \\ g *)
  
  intros b Hb.
  apply ut_eta_abs_inv in Hb.
  destruct Hb as [ Hb | (w & ? & Hw) ]; subst.

  apply ut_beta_app_inv in H.
  destruct H as [ (f & ? & ?)
              | [ (h & ? & H)
                | (w & ? & H) ] ]; subst.

  exists (\\ (f[£ pos0])).
  split.
  apply ut_eta_rt_refl.
  right.
  apply ut_lift_inj.
  rewrite H.
  unfold ut_lift; simpl.
  f_equal.
  unfold ut_reduce.
  rewrite ut_ren_subst.
  rewrite <- ut_subst_id at 1.
  apply ut_subst_occ_ext.
  clear IH.
  intros p Hp; pos_inv p; simpl.
  cbv; auto.
  pos_inv p; cbv; auto.
  apply f_equal with (f := fun f => ut_occ f pos0) in H.
  simpl in H.
  rewrite ut_occ_lift_fst in H.
  exfalso; revert Hp; rewrite H; apply lt_irrefl.

  apply ut_beta_lift_inv in H.
  destruct H as (g & ? & H); subst.
  exists g.
  split.
  apply ut_eta_rt_redex, ut_eta_rt_refl.
  left; auto.

  apply ut_beta_var_inv in H.
  destruct H.

  destruct (IH _ Hw) as (h & H1 & H2).
  exists (\\ h); split.
  apply ut_eta_rt_abs; auto.
  destruct H2 as [ H2 | H2 ].
  left; constructor 4; auto.
  subst; right; auto.
Qed.

Local Lemma ut_beta_or_eq_eta_rt_comm n : commute (@ut_beta n cup2 eq) (@ut_eta_rt n).
Proof.
  apply square_inc with (1 := @clos_refl_trans_idem _ _).
  apply square_comm, square_rt.
  apply square_comm, square_cup.
  intros ? ? ?; apply ut_beta_eta_square.
  apply square_comm, square_inc with (T := eq); auto.
  apply square_comm, square_inc with (1 := @clos_refl_trans_inc _ _).
  apply square_eq.
Qed.

Theorem ut_beta_rt_eta_rt_commute n : commute (@ut_beta_rt n) (@ut_eta_rt n).
Proof.
  apply square_dec with (R := clos_refl_trans _ (@ut_beta _ cup2 eq)).
  apply clos_refl_trans_mono; auto.
  apply square_comm.
  apply square_inc with (T := clos_refl_trans _ (@ut_beta _ cup2 eq)).
  apply inc2_trans with (2 := @clos_refl_trans_idem _ _), clos_refl_trans_mono.
  intros ? ? [ | [] ]; [ constructor 1 | constructor 2 ]; auto.
  apply square_comm, square_rt.
  apply ut_beta_or_eq_eta_rt_comm.
Qed. 

Theorem ut_beta_eta_confluent n : confluent (@ut_beta_eta n).
Proof.
  apply Hindley_Rosen.
  apply ut_beta_confluent.
  apply ut_eta_confluent.
  apply ut_beta_rt_eta_rt_commute.
Qed.
