(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import Arith Omega List.

Require Import utils cr pos vec rec_alg cl.
Require Import uterm uterm_beta uterm_beta_eq uterm_normal.
Require Import uterm_bool uterm_nat.
Require Import uterm_fixpoint uterm_total.

Set Implicit Arguments.

Fixpoint uterm_clterm n (f : uterm n) : clterm n :=
  match f with
    | \\ f  => cl_abs (uterm_clterm f)
    | f @ a => uterm_clterm f ° uterm_clterm a
    | £ v   => µ v
  end.
  
Local Notation " '[[' f ']]' " := (uterm_clterm f).

Definition cl_E n : clterm n := [[ \\ \\ (£ pos1 @ £ pos0) ]].

Fact cl_E_prop_1 n (x : clterm n) : cl_E n ° x ~[cl0] (cl_S _ ° (cl_K _ ° x)) ° cl_I _.
Proof.
  unfold cl_E.
  apply cl_eq_trans with (1 := cl_abs_prop _ _).
  apply cl_eq_refl; auto.
Qed.

Fact cl_E_prop_2 n (x y : clterm n) : cl_E n ° x ° y ~[cl0] x ° y.
Proof.
  apply cl_eq_trans with ((cl_S _ ° (cl_K _ ° x)) ° cl_I _ ° y).
  apply cl_eq_app; auto.
  apply cl_E_prop_1.
  apply cl_eq_trans with (1 := cl_S_prop _ _ _).
  apply cl_eq_app.
  apply cl_K_prop.
  apply cl_I_prop.
Qed.

Fact cl_E_prop_3 n (x : clterm n) : cl_abs (cl_lift x ° µ pos0) ~[cl0] cl_E _ ° x.
Proof.
  apply cl_eq_sym.
  apply cl_eq_trans with (1 := cl_E_prop_1 _).
  apply cl_eq_refl.
  rewrite cl_abs_app.
  2: simpl; omega.
  rewrite cl_abs_pos0.
  rewrite cl_lift_rev_prop0; auto.
  apply cl_occ_lift_pos0.
Qed.

Fact ut_clterm_ren n (f : uterm n) m (s : pos n -> pos m) : [[ ut_ren f s ]] = cl_subst [[ f ]] (fun p => µ (s p)).
Proof.
  revert m s.
  induction f as [ n f Hf | n f Hf a Ha | n p ]; intros m s; simpl; auto.
  rewrite cl_subst_abs; f_equal.
  rewrite Hf.
  apply cl_subst_ext.
  intros p Hp; pos_inv p; auto.
  f_equal; auto.
Qed.

Fact ut_clterm_subst n (f : uterm n) m (s : pos n -> uterm m) : [[ ut_subst f s ]] = cl_subst [[ f ]] (fun p => [[s p]]).
Proof.
  revert m s.
  induction f as [ n f Hf | n f Hf a Ha | n p ]; intros m s; simpl; auto.
  rewrite cl_subst_abs; f_equal.
  rewrite Hf; apply cl_subst_ext.
  intros p Hp; pos_inv p; auto.
  unfold uts_lift; simpl.
  unfold ut_lift; rewrite ut_clterm_ren.
  apply cl_subst_ext; auto.
  f_equal; auto.
Qed.

Inductive cl1 n : clterm n -> clterm n -> Prop :=
  | in_cl1_eq_K : cl1 (cl_K n) (cl_abs (cl_abs (cl_K _ ° µ pos1 ° µ pos0)))
  | in_cl1_eq_S : cl1 (cl_S n) (cl_abs (cl_abs (cl_abs (cl_S _ ° µ pos2 ° µ pos1 ° µ pos0))))
  .
  
Definition cl01 n := @cl0 n cup2 @cl1 n.

Fact cl0_inc_cl01 n (x y : clterm n) : x ~[cl0] y -> x ~[cl01] y.
Proof.
  apply cl_eq_mono; left; auto.
Qed.

Fact cl1_inc_cl01 n (x y : clterm n) : x ~[cl1] y -> x ~[cl01] y.
Proof.
  apply cl_eq_mono; right; auto.
Qed.
  
Fact cl01_K n (x : clterm n) : cl_K _ ° x ~[cl01] cl_abs (cl_K _ ° cl_lift x ° µ pos0).
Proof.
  apply cl_eq_trans with (cl_abs (cl_abs (cl_K _ ° µ pos1 ° µ pos0)) ° x).
  apply cl_eq_app; auto.
  apply cl1_inc_cl01; constructor 1; constructor 1.
  apply cl0_inc_cl01.
  apply cl_eq_trans with (1 := cl_abs_prop _ _).
  apply cl_eq_refl.
  unfold cl_reduce.
  simpl;f_equal.
  rewrite plus_comm; simpl.
  do 2 f_equal.
  generalize (cl_occ_lift_pos0 x); simpl; intros E; rewrite E.
  do 2 f_equal.
  symmetry; apply cl_lift_rev_prop0.
Qed.

Fact cl01_S n (x y : clterm n) : cl_S _ ° x ° y ~[cl01] cl_abs (cl_S _ ° cl_lift x ° cl_lift y ° µ pos0).
Proof.
  apply cl_eq_trans with ((cl_abs (cl_abs (cl_abs (cl_S _ ° µ pos2 ° µ pos1 ° µ pos0)))) ° x ° y).
  apply cl_eq_app; auto.
  apply cl_eq_app; auto.
  apply cl1_inc_cl01; constructor 1; constructor 2.
  apply cl_eq_trans with (cl_reduce (cl_abs (cl_abs (cl_S (S (S (S n))) ° µ (pos2) ° µ (pos1) ° µ (pos0)))) x ° y).
  apply cl_eq_app; auto.
  apply cl0_inc_cl01.
  apply cl_eq_trans with (1 := cl_abs_prop _ _); auto.
  unfold cl_reduce.
  do 2 rewrite cl_subst_abs.
  apply cl0_inc_cl01.
  apply cl_eq_trans with (1 := cl_abs_prop _ _); auto.
  apply cl_eq_refl.
  unfold cl_reduce.
  rewrite cl_subst_abs, cl_subst_comp.
  f_equal.
  simpl.
  repeat f_equal.
  do 2 rewrite cl_subst_comp.
  apply cl_subst_ext; intros; auto.
Qed.

Inductive cl1_0 n : clterm n -> clterm n -> Prop :=
  | in_cl1_0_eq_K : forall x,   cl1_0 (cl_E _ ° (cl_K _ ° x)) (cl_K _ ° x)
  | in_cl1_0_eq_S : forall x y, cl1_0 (cl_E _ ° (cl_S _ ° x ° y)) (cl_S _ ° x ° y)
  .
  
Fact cl1_0_K n x : cl_E n ° (cl_K _ ° x) ~[cl01] cl_K _ ° x.
Proof.
  apply cl_eq_sym.
  apply cl_eq_trans with (1 := cl01_K _).
  apply cl0_inc_cl01.
  apply cl_eq_trans with (2 := cl_E_prop_3 _).
  apply cl_eq_refl; auto.
Qed.

Fact cl1_0_S n x y : cl_E n ° (cl_S _ ° x ° y) ~[cl01] cl_S _ ° x ° y.
Proof.
  apply cl_eq_sym.
  apply cl_eq_trans with (1 := cl01_S _ _).
  apply cl0_inc_cl01.
  apply cl_eq_trans with (2 := cl_E_prop_3 _).
  apply cl_eq_refl; auto.
Qed.

Fact cl1_0_inc_cl01 n (x y : clterm n) : cl1_0 x y -> x ~[cl01] y.
Proof.
  induction 1.
  apply cl1_0_K.
  apply cl1_0_S.
Qed.

Definition cl01_0 n := @cl0 n cup2 @cl1_0 n.

Fact cl0_cl01_0 n (x y : clterm n) : x ~[cl0] y -> x ~[cl01_0] y.
Proof.
  apply cl_eq_mono; left; auto.
Qed.

Fact cl1_0_cl01_0 n (x y : clterm n) : x ~[cl1_0] y -> x ~[cl01_0] y.
Proof.
  apply cl_eq_mono; right; auto.
Qed.

Fact cl01_0_inc_cl01 n : forall (x y : clterm n), x ~[cl01_0] y -> x ~[cl01] y.
Proof.
  apply cl_eq_closure.
  intros x y [ H | H ].
  apply cl0_inc_cl01.
  constructor 1; auto.
  apply cl1_0_inc_cl01; auto.
Qed.

Fact cl01_0_E_prop_0 n (t : clterm (S n)) : cl_abs t ~[cl01_0] cl_E _ ° (cl_abs t).
Proof.
  destruct t.
  rewrite cl_abs_pos0; auto.
  apply cl_eq_sym; constructor 1; constructor 2; constructor 1.
  rewrite cl_abs_pos0; auto.
  apply cl_eq_sym; constructor 1; constructor 2; constructor 1.
  case_eq (cl_occ (t1 ° t2) pos0).
  intros ?.
  rewrite cl_abs_pos0; auto.
  apply cl_eq_sym; constructor 1; constructor 2; constructor 1.
  intros k Hk.
  rewrite cl_abs_app.
  2: simpl in Hk; omega.
  apply cl_eq_sym; constructor 1; constructor 2; constructor 2.
  pos_inv p; simpl.
  apply cl_eq_sym; constructor 1; constructor 2; constructor 2.
  apply cl_eq_sym; constructor 1; constructor 2; constructor 1.
Qed.

Fact cl01_0_E_prop_1 n (t : clterm (S n)) : cl_E _ ° (cl_abs t) ~[cl01_0] cl_abs (cl_lift (cl_abs t) ° µ pos0).
Proof.
  apply cl_eq_sym, cl0_cl01_0, cl_E_prop_3.
Qed.

Fact cl01_0_E_prop_2 n : cl_E n ° cl_E _ ~[cl01_0] cl_E _.
Proof.
  apply cl_eq_sym, cl01_0_E_prop_0.
Qed.

Fact cl01_0_E_prop_3 n x : cl_E n ° (cl_E _ ° x) ~[cl01_0] cl_E _ ° x.
Proof.
  apply cl_eq_trans with (cl_E _ ° (cl_abs (cl_lift x ° µ pos0))).
  apply cl_eq_app; auto.
  apply cl0_cl01_0, cl_eq_sym, cl_E_prop_3.
  apply cl_eq_sym.
  apply cl_eq_trans with (2 := cl01_0_E_prop_0 _).
  apply cl0_cl01_0, cl_eq_sym, cl_E_prop_3.
Qed.

Fixpoint clterm_uterm n (f : clterm n) : uterm n :=
  match f with 
    | cl_K _ => \\ \\ £ pos1
    | cl_S _ => \\ \\ \\ (£ pos2 @ £ pos0 @ (£ pos1 @ £ pos0))
    | f ° a  => clterm_uterm f @ clterm_uterm a
    | µ p    => £ p
  end.
  
Local Notation " '[|' f '|]' " := (clterm_uterm f).

Fact clterm_ut_app n (f a : clterm n) : [| f ° a |] = [|f|] @ [|a|].
Proof. auto. Qed. 

Require Import uterm_beta_eq.

Fact clterm_ut_ren n f m r : @ut_ren n [|f|] m r = [| cl_subst f (fun p => µ (r p)) |].
Proof.
  induction f; simpl; f_equal; auto.
Qed.

Theorem cl01_beta_eq n (u v : clterm n) : u ~[cl01] v -> [|u|] ~b [|v|].
Proof.
  induction 1.
  destruct H.
  induction H.
  
  simpl.
  generalize [|x|] [|y|]; clear x y; intros x y.
  apply ut_beta_rt_eq.
  ut_beta_rt_step.
  unfold ut_reduce; simpl.
  unfold uts_lift; simpl.
  unfold uts_first; simpl.
  ut_beta_rt_step.
  rewrite ut_reduce_lift.
  apply ut_beta_rt_refl.
  
  simpl.
  generalize [|x|] [|y|] [|z|]; clear x y z; intros x y z.
  apply ut_beta_rt_eq.
  do 2 (ut_beta_rt_step;
  unfold ut_reduce; simpl; 
  unfold uts_lift; simpl; 
  unfold uts_first; simpl).
  unfold ut_lift.
  rewrite ut_subst_ren;simpl.
  rewrite ut_subst_ren;simpl.
  rewrite <- ut_lift_subst_eq.
  change (ut_ren y (pos_nxt n)) with (ut_lift y).
  ut_beta_rt_step.
  repeat rewrite ut_reduce_app.
  do 2 rewrite ut_reduce_lift.
  rewrite ut_reduce_pos0.
  apply ut_beta_rt_refl.
  
  induction H.
  
  (* we use a non terminating cbv algorithm to compute the normal form,
     given a bound on the number of reduction. The minimal bound is 
     easy to find by dichotomy *)
  
  simpl; apply ut_beta_eq_sym, ut_beta_rt_eq.
  apply ut_beta_cbv; exists 36; reflexivity.
  
  simpl; apply ut_beta_eq_sym, ut_beta_rt_eq.
  apply ut_beta_cbv; exists 85; reflexivity.
  
  apply ut_beta_eq_refl.
  apply ut_beta_eq_sym; auto.
  apply ut_beta_eq_trans with ([|y|]); auto.
  simpl; apply ut_beta_eq_app; auto; apply ut_beta_eq_refl.  
  simpl; apply ut_beta_eq_app; auto; apply ut_beta_eq_refl.
Qed.

Fact cl_E_beta_eq n : [|cl_E n|] ~b \\ \\ (£ pos1 @ £ pos0).
Proof.
  unfold cl_E; simpl.
  apply ut_beta_rt_eq.
  apply ut_beta_cbv; exists 30; reflexivity.
Qed.

Fact Wext_lift_beta_eq n : forall y z : uterm n, ut_lift y @ £ pos0 ~b ut_lift z @ £ pos0 -> [|cl_E _|] @ y ~b [|cl_E _|] @ z.
Proof.
  intros y z H.
  apply ut_beta_eq_trans with ((\\ \\ (£ pos1 @ £ pos0)) @ y).
  apply ut_beta_eq_app; [ apply cl_E_beta_eq | apply ut_beta_eq_refl ].
  apply ut_beta_eq_trans with ((\\ \\ (£ pos1 @ £ pos0)) @ z).
  2: apply ut_beta_eq_sym, ut_beta_eq_app; [ apply cl_E_beta_eq | apply ut_beta_eq_refl ].
  apply ut_beta_eq_trans with (\\ (ut_lift y @ £ pos0)).
  apply ut_beta_rt_eq; ut_beta_rt_step;apply ut_beta_rt_refl.
  apply ut_beta_eq_trans with (\\ (ut_lift z @ £ pos0)).
  2: apply ut_beta_eq_sym, ut_beta_rt_eq; ut_beta_rt_step;apply ut_beta_rt_refl.
  apply ut_beta_eq_abs; auto.
Qed.

(* We need a free variable to show that result *)

Fact Wext_beta_eq n : forall y z : uterm n, (forall u, ut_lift y @ u ~b ut_lift z @ u) -> [|cl_E _|] @ ut_lift y ~b [|cl_E _|] @ ut_lift z.
Proof.
  intros a b H; apply Wext_lift_beta_eq.
  specialize (H (£ pos0)).
  apply ut_beta_eq_ren with (r := fun p => match p with pos_fst _ => pos0 | pos_nxt _ p => pos_nxt _ (pos_nxt _ p) end) in H.
  simpl in H.
  revert H; eqgoal; f_equal; f_equal;
  unfold ut_lift; do 2 rewrite <- ut_ren_comp; apply ut_ren_ext; cbv; auto.
Qed.

Fact cl_subst_ut_ren n (f : uterm n) m (r : pos n -> pos m) :
   cl_subst [[f]] (fun p => µ (r p)) = [[ ut_ren f r ]].
Proof.
  revert m r; induction f as [ n f Hf | n f Hf a Ha | n v ]; intros m r; simpl; auto.
  rewrite cl_subst_abs; f_equal.
  rewrite <- Hf.
  apply cl_subst_ext; intros p ?; pos_inv p; auto.
  f_equal; auto.
Qed.

Fact cl_subst_cls_lift n (f : uterm (S n)) m (r : pos n -> pos m) : 
     cl_subst [[ f ]] (cls_lift (fun p => µ (r p))) 
   = [[ ut_ren f (pm_lift r) ]].
Proof.
  rewrite <- cl_subst_ut_ren.
  apply cl_subst_ext.
  intros p ?; pos_inv p; auto.
Qed.

(* That proof should be cleaned up *)

Fact ut_cl_ut_eq n (f : uterm n) : [| [[ f ]] |] ~b f.
Proof.
  induction f as [ n f Hf | n f Hf a Ha | n v ].
  2: simpl; apply ut_beta_eq_app; auto.
  2: apply ut_beta_eq_refl.

  simpl.
  generalize (cl_lift_abs_prop [[f]]); intros H.
  apply cl0_inc_cl01, cl01_beta_eq in H.
  generalize (ut_beta_eq_trans H Hf); intros H1.
  simpl clterm_uterm in H1.
  rewrite cl_subst_abs in H1.
  rewrite cl_subst_cls_lift in H1.  
  revert H1; set (f' := ut_ren f (pm_lift (pos_nxt n))); intros H1.
  generalize (cl_E_beta_eq n); intros H2.
  generalize (ut_beta_eq_app H2 (ut_beta_eq_refl [|cl_abs [[f]]|])); intros H3.
  generalize (cl01_0_E_prop_0 [[f]]); intros H4.
  apply cl01_0_inc_cl01, cl01_beta_eq in H4.
  rewrite clterm_ut_app in H4.
  apply ut_beta_eq_trans with (1 := H4),
        ut_beta_eq_trans with (1 := H3).
  apply ut_beta_eq_trans with (\\ ([|cl_abs [[f']]|] @ £ (pos0))).
  2:  apply ut_beta_eq_abs; auto.
  apply ut_beta_rt_eq.
  ut_beta_rt_step.
  apply ut_beta_rt_refl_eq.
  unfold f'.
  unfold ut_reduce, uts_first; simpl.
  do 2 f_equal.
  unfold uts_lift; simpl.
  unfold ut_lift.
  rewrite clterm_ut_ren, cl_subst_abs.
  do 2 f_equal.
  rewrite <- cl_subst_ut_ren.
  apply cl_subst_ext.
  intros p ?; pos_inv p; auto.
Qed.

(** every lambda term is beta equivalent to a combinator
 ** 
 ** We intend to use this to build a term (u : uterm n)
 ** st that 
 **
 **    forall t : uterm n, exists k, t ~b u @ ut_nat n k 
 ** 
 ** hence u is a lambda enumerator
 **) 

Theorem cl_ut_surj n (f : uterm n) : { g | f ~b [|g|] }.
Proof.
  exists [[f]].
  apply ut_beta_eq_sym, ut_cl_ut_eq.
Qed.

Fixpoint cl2nat n (f : clterm n) : nat :=
  match f with
    | cl_S _ => 0
    | cl_K _ => 1
    | µ p    => 2 + pos2nat p
    | f ° a  => 2 + n + nat_pair (cl2nat f) (cl2nat a)
  end.
  
Fact cl2nat_inj n (f g : clterm n) : cl2nat f = cl2nat g -> f = g.
Proof.
  revert g; induction f as [ | | f Hf a Ha | v ]; intros [ | | g b | w ]; simpl; try discriminate 1; auto.
  intros H; injc H.
  apply plus_reg_l, nat_pair_inj in H.
  rewrite (Hf g), (Ha b); auto; omega.
  intros H; exfalso; generalize (pos2nat_prop w); omega.
  intros H; exfalso; generalize (pos2nat_prop v); omega.
  intros H; injc H.
  apply pos2nat_inj in H; f_equal; auto.
Qed.

Definition analyze n (k : nat) : { k = 0 } 
                               + { k = 1 } 
                               + { p : pos n | k = 2 + pos2nat p } 
                               + { x : nat & { y | k = 2+n+nat_pair x y } }.
Proof.
  destruct k as [ | [ | k ] ]; auto.
  destruct (le_lt_dec n k) as [ H | H ].
  right.
  exists (nat_fst (k-n)), (nat_snd (k-n)).
  rewrite nat_pair_fst_snd; omega.
  left; right.
  exists (nat2pos H).
  rewrite pos2nat_nat2pos; auto.
Qed.

Definition nat2cl_rec n (k : nat) : { f : clterm n | cl2nat f = k }.
Proof.
  induction k as [ k IHk ] using (well_founded_induction_type lt_wf).
  destruct (analyze n k) as [ [ [ ? | ? ] | (p & Hp) ] | (x & y & Hxy) ]; subst.

  exists (cl_S _); auto.
  exists (cl_K _); auto.
  exists (µ p); auto.
  destruct (IHk x) as (f & Hf).
  apply le_lt_trans with (1 := nat_pair_le_lft x y); omega.
  destruct (IHk y) as (a & Ha).
  apply le_lt_trans with (1 := nat_pair_le_rt x y); omega.
  exists (f ° a); subst; auto.
Qed.

Definition nat2cl n k := proj1_sig (nat2cl_rec n k).
Fact nat2cl_prop n k : cl2nat (nat2cl n k) = k.
Proof.
  apply (proj2_sig (nat2cl_rec n k)).
Qed.

Fact cl2nat_prop n f : nat2cl n (cl2nat f) = f.
Proof.
  apply cl2nat_inj; rewrite nat2cl_prop; auto.
Qed.

Fact nat2cl_prop_0 n : nat2cl n 0 = cl_S n.
Proof.
  apply cl2nat_inj; rewrite nat2cl_prop; auto.
Qed.

Fact nat2cl_prop_1 n  : nat2cl n 1 = cl_K n.
Proof.
  apply cl2nat_inj; rewrite nat2cl_prop; auto.
Qed.

Fact nat2cl_prop_p n p (H : p < n) : nat2cl n (2+p) = µ (nat2pos H).
Proof.
  apply cl2nat_inj; simpl; rewrite nat2cl_prop, pos2nat_nat2pos; auto.
Qed.

Fact nat2cl_prop_app n k : nat2cl n (2+n+k) = nat2cl n (nat_fst k) ° nat2cl n (nat_snd k).
Proof.
  apply cl2nat_inj; simpl; repeat rewrite nat2cl_prop; do 3 f_equal.
  symmetry; apply nat_pair_fst_snd.
Qed.

(*

We implement the following fixpoint as a lambda term 

Definition phi n (h : nat -> clterm n) k : clterm n :=
  match k with 
    | 0       => cl_S _
    | 1       => cl_K _
    | S (S k) => match le_lt_dec n k with
                   | left _   => h (nat_fst (k - n)) ° h (nat_snd (k - n))
                   | right H  => µ (nat2pos H)
                 end
  end.
  
Arguments phi {n} h k /.

Fact phi_fix n k : nat2cl n k = phi (nat2cl n) k.
Proof.
  induction k as [ k IHk ] using (well_founded_induction_type lt_wf).
  destruct (analyze n k) as [ [ [ ? | ? ] | (p & Hp) ] | (x & y & Hxy) ]; subst; simpl.
  apply nat2cl_prop_0.
  apply nat2cl_prop_1.
  destruct (le_lt_dec n (pos2nat p)).
  exfalso; generalize (pos2nat_prop p); omega.
  apply nat2cl_prop_p.
  destruct (le_lt_dec n (n + nat_pair x y)).
  change (S (S (n+nat_pair x y))) with (2+n+nat_pair x y).
  rewrite nat2cl_prop_app; do 3 f_equal; omega.
  exfalso; omega.
Qed.

*)

Section phi.

  Variable (n : nat).
  
  Let vt1 : vec (uterm n) _ := [|cl_S _|] ## [|cl_K _|] ## vec_set_pos (fun p => £ p).
  
  Let t1 := ut_nth n @ ut_list (vec_list vt1).
 
  Let t1_0 : t1 @ ut_nat n 0 -b>> [|cl_S _|].
  Proof.
    unfold t1.
    assert (0 < S (S n)) as H0; [ omega | ].
    apply ut_beta_rt_trans with (1 := ut_nth_prop _ (lt_0_Sn _)).
    apply ut_beta_rt_refl_eq; auto.
  Qed.
  
  Let t1_1 : t1 @ ut_nat n 1 -b>> [|cl_K _|].
  Proof.
    unfold t1.
    assert (1 < S (S n)) as H0; [ omega | ].
    apply ut_beta_rt_trans with (1 := ut_nth_prop _ H0).
    apply ut_beta_rt_refl_eq; auto.
  Qed.
  
  Let t1_2 p : t1 @ ut_nat n (2+pos2nat p) -b>> £ p.
  Proof.
    unfold t1.
    assert (2 + pos2nat p < S (S n)) as H0.
      generalize (pos2nat_prop p); omega.   
    apply ut_beta_rt_trans with (1 := ut_nth_prop _ H0).
    unfold vt1; simpl.
    apply ut_beta_rt_refl_eq; auto.
    rewrite vec_get_pos, vec_pos_set; auto.
  Qed.
  
  Let t2 : uterm n := \\ \\ \\ (ut_ite _ @ (ut_leq _ @ £ pos0 @ ut_nat _ (S n)) @ (£ pos2 @ £ pos0) @ (£ pos1 @ £ pos0)).
  
  Let t2_0 k x y : t2 @ x @ y @ k -b>> ut_ite _ @ (ut_leq _ @ k @ ut_nat _ (S n)) @ (x @ k) @ (y @ k).
  Proof.
    unfold t2.
    ut_beta_rt_step.
    unfold ut_reduce.
    do 2 rewrite ut_subst_abs.
    repeat rewrite ut_subst_app.
    repeat rewrite ut_subst_var.
    rewrite ut_ite_subst, ut_leq_subst, ut_nat_subst.
    unfold uts_lift; simpl.
    unfold uts_first; simpl.
    unfold ut_lift at 3; simpl.
    ut_beta_rt_step.
    unfold ut_reduce.
    do 1 rewrite ut_subst_abs.
    repeat rewrite ut_subst_app.
    repeat rewrite ut_subst_var.
    rewrite ut_ite_subst, ut_leq_subst, ut_nat_subst.
    rewrite <- ut_lift_subst_2, ut_subst_lift_first.
    unfold uts_lift; simpl.
    unfold uts_first; simpl.
    ut_beta_rt_step.
    unfold ut_reduce.
    repeat rewrite ut_subst_app.
    repeat rewrite ut_subst_var.
    rewrite ut_ite_subst, ut_leq_subst, ut_nat_subst.
    repeat rewrite ut_subst_lift_first.
    unfold uts_first; simpl.
    apply ut_beta_rt_refl.
  Qed.
  
  Let t2_1 k x y : k < S (S n) -> t2 @ x @ y @ ut_nat _ k -b>> x @ ut_nat _ k.
  Proof.
    intros H.
    apply ut_beta_rt_trans with (1 := t2_0 _ _ _).
    apply ut_ite_true_beta_prop.
    apply ut_leq_prop_true; omega.
  Qed.
  
  Let t2_2 k x y : S (S n) <= k -> t2 @ x @ y @ ut_nat _ k -b>> y @ ut_nat _ k.
  Proof.
    intros H.
    apply ut_beta_rt_trans with (1 := t2_0 _ _ _).
    apply ut_ite_false_beta_prop.
    apply ut_leq_prop_false; omega.
  Qed.
  
  Let t3 : uterm n := \\ \\ ( (£ pos1 @ (ut_nat_fst _ @ (ut_sub _ @ £ pos0 @ ut_nat _ (S (S n)))))
                            @ (£ pos1 @ (ut_nat_snd _ @ (ut_sub _ @ £ pos0 @ ut_nat _ (S (S n))))) ).
                            
  Let t3_0 f k : t3 @ f @ ut_nat _ k -b>> (f @ ut_nat _ (nat_fst (k - S (S n)))) 
                                        @ (f @ ut_nat _ (nat_snd (k - S (S n)))).
  Proof.
    unfold t3.
    
    ut_beta_rt_step.
    unfold ut_reduce.
    rewrite ut_subst_abs.
    repeat rewrite ut_subst_app.
    repeat rewrite ut_subst_var.
    rewrite ut_nat_fst_subst, ut_nat_snd_subst.
    repeat rewrite ut_sub_subst.
    repeat rewrite ut_nat_subst.
    unfold uts_lift; simpl.
    unfold uts_first; simpl.
    
    ut_beta_rt_step.
    unfold ut_reduce.
    repeat rewrite ut_subst_app.
    repeat rewrite ut_subst_lift_first.
    repeat rewrite ut_subst_var.
    rewrite ut_nat_fst_subst, ut_nat_snd_subst.
    repeat rewrite ut_sub_subst.
    repeat rewrite ut_nat_subst.
    unfold uts_first; simpl.
    apply ut_beta_rt_app; apply ut_beta_rt_rt.
    apply ut_beta_rt_trans with (2 := ut_nat_fst_prop _ _), ut_beta_rt_rt, ut_sub_prop.
    apply ut_beta_rt_trans with (2 := ut_nat_snd_prop _ _), ut_beta_rt_rt, ut_sub_prop.
  Qed.

  Let t4 := \\ (ut_lift t2 @ ut_lift t1 @ (ut_lift t3 @ £ pos0)).
  
  Let t4_0 f : t4 @ f -b>> t2 @ t1 @ (t3 @ f).
  Proof.
    unfold t4.
    ut_beta_rt_step.
    unfold ut_reduce.
    repeat rewrite ut_subst_app.
    repeat rewrite ut_subst_lift_first.
    rewrite ut_subst_var.
    apply ut_beta_rt_refl.
  Qed.
  
  Let ut_phi := ut_fix_turing _ @ t4.
  
  Let ut_phi_0 : ut_phi -b>> t2 @ t1 @ (t3 @ ut_phi).
  Proof.
    unfold ut_phi at 1.
    apply ut_beta_rt_trans with (1 := ut_fix_turing_beta_reduce _).
    apply ut_beta_rt_trans with (1 := t4_0 _).
    apply ut_beta_rt_refl.
  Qed.
   
  Let ut_phi_1 k : ut_phi @ ut_nat _ k -b>> t2 @ t1 @ (t3 @ ut_phi) @ ut_nat _ k.
  Proof. apply ut_beta_rt_lft, ut_phi_0. Qed.
  
  (* the next three results are the base case of induction *)
  
  Let ut_phi_prop_0 : ut_phi @ ut_nat _ 0 -b>> [|cl_S _|].
  Proof.
    apply ut_beta_rt_trans with (1 := ut_phi_1 _).
    assert (0 < S (S n)) as H0; try omega.
    apply ut_beta_rt_trans with (1 := t2_1 _ _ H0); auto.
  Qed.
  
  Let ut_phi_prop_1 : ut_phi @ ut_nat _ 1 -b>> [|cl_K _|].
  Proof.
    apply ut_beta_rt_trans with (1 := ut_phi_1 _).
    assert (1 < S (S n)) as H0; try omega.
    apply ut_beta_rt_trans with (1 := t2_1 _ _ H0); auto.
  Qed.
  
  Let ut_phi_prop_2 p : ut_phi @ ut_nat _ (2+pos2nat p) -b>> £ p.
  Proof.
    apply ut_beta_rt_trans with (1 := ut_phi_1 _).
    assert (2 + pos2nat p < S (S n)) as H0.
      generalize (pos2nat_prop p); omega.   
    apply ut_beta_rt_trans with (1 := t2_1 _ _ H0); auto.
  Qed.
  
  (* this is the recursive call *)
  
  Let ut_phi_prop_3 k : S (S n) <= k -> ut_phi @ ut_nat _ k -b>> (ut_phi @ ut_nat _ (nat_fst (k - S (S n))))
                                                               @ (ut_phi @ ut_nat _ (nat_snd (k - S (S n)))).
  Proof.
    intros H.
    apply ut_beta_rt_trans with (1 := ut_phi_1 _).
    apply ut_beta_rt_trans with (1 := t2_2 _ _ H); auto.
  Qed.
  
  Let ut_phi_beta_prop k : ut_phi @ ut_nat _ k -b>> [| nat2cl _ k |].
  Proof.
    induction k as [ k IHk ] using (well_founded_induction_type lt_wf).
    destruct (analyze n k) as [ [ [ ? | ? ] | (p & Hp) ] | (x & y & Hxy) ]; subst.
    
    rewrite nat2cl_prop_0; apply ut_phi_prop_0.
    
    rewrite nat2cl_prop_1; apply ut_phi_prop_1.
    
    rewrite nat2cl_prop_p with (H := pos2nat_prop p).
    rewrite nat2pos_pos2nat.
    apply ut_phi_prop_2.
    
    assert (S (S n) <= 2+n+nat_pair x y) as H0; try omega.
    apply ut_beta_rt_trans with (1 := ut_phi_prop_3 H0).
    rewrite nat2cl_prop_app.
    simpl clterm_uterm.
    cutrewrite (2 + n + nat_pair x y - S (S n) = nat_pair x y); try omega.
    apply ut_beta_rt_app; apply IHk.
    apply le_lt_trans with (1 := nat_fst_le _); omega.
    apply le_lt_trans with (1 := nat_snd_le _); omega.
  Qed.
  
  Definition ut_universal := ut_phi.
  
  Theorem ut_universal_prop k : ut_universal @ ut_nat _ k ~b [| nat2cl _ k |].
  Proof.
    apply ut_beta_rt_eq, ut_phi_beta_prop.
  Qed.
   
End phi.

(* For any n, there exists a universal lambda term (ut_universal n : uterm n) which enumerates 
   every lambda term in uterm n upto beta equivalence ie the map
   
      k --> ut_universal n @ ut_nat n k 
      
   is surjective on the beta-equivalence classes of lambda terms
   
   Has explained above, it is build via a detour through combinatory logic 
*)

Theorem ut_has_universal n (t : uterm n) : { k | t ~b ut_universal n @ ut_nat _ k }.
Proof.
  destruct (cl_ut_surj t) as (f & Hf).
  exists (cl2nat f).
  apply ut_beta_eq_sym.
  apply ut_beta_eq_trans with (1 := ut_universal_prop n _).
  apply ut_beta_eq_sym.
  rewrite cl2nat_prop; auto.
Qed.

Check ut_has_universal.
Print Assumptions ut_has_universal.


