(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import List Arith Omega.

Require Import utils pos vec rec_alg.

Require Import uterm.

Set Implicit Arguments.

Section uterm_enc.

  Notation "'<<' x ; y '>>'" := (nat_pair x y).

  Definition uterm_enc : forall k, uterm k -> nat.
  Proof.
    induction 1 as [ k f Hf | k f Hf a Ha | k p ].
    exact (<< k ; << 1 ; Hf >> >>).
    exact (<< k ; << 2 ; << Hf ; Ha >> >> >>).
    exact (<< k ; << 3 ; pos2nat p >> >>).
  Defined.
  
  (* we define the decoding algorithm together with its correctness proof *)
  
  Definition uterm_dec_rec : forall n k, { t : uterm k | forall x, @uterm_enc k x = n -> t = x }. 
  Proof.
    induction n as [ n IHn ] using (well_founded_induction_type lt_wf); intros k.
    destruct (eq_nat_dec (nat_fst n) k) as [ H | C ].
    
    Focus 2.
    exists (\\ £ pos0); intros [ f | f a | p ]; simpl; intros; subst n;
    rewrite nat_fst_pair_eq in C; destruct C; auto.
    
    remember (nat_fst (nat_snd n)) as p.
    destruct p as [ | [ | [ | [ | p ] ] ] ].
    
    exists (\\ £ pos0); intros [ f | f a | v ]; simpl; intros; subst n;
    rewrite nat_snd_pair_eq, nat_fst_pair_eq in Heqp; discriminate Heqp.
    
    Focus 4.
    exists (\\ £ pos0); intros [ f | f a | v ]; simpl; intros; subst n;
    rewrite nat_snd_pair_eq, nat_fst_pair_eq in Heqp; discriminate Heqp.
    
    destruct (IHn (nat_snd (nat_snd n))) with (k := S k) as (t & Ht).
    rewrite <- nat_pair_fst_snd.
    rewrite <- (nat_pair_fst_snd (nat_snd n)) at 2.
    rewrite H, <- Heqp.
    generalize (nat_snd (nat_snd n)); intros x.
    apply le_lt_trans with (<<1;x>>).
    apply nat_pair_le_rt.
    apply nat_pair_lt_rt.
    apply nat_pair_lt_lft; auto.
    exists (\\t); intros [ g | g a | v ] H'; subst n.
    f_equal; apply (Ht g); simpl.
    do 2 rewrite nat_snd_pair_eq; auto.
    simpl in Heqp; rewrite nat_snd_pair_eq, nat_fst_pair_eq in Heqp; discriminate Heqp.
    simpl in Heqp; rewrite nat_snd_pair_eq, nat_fst_pair_eq in Heqp; discriminate Heqp.
    
    destruct (IHn (nat_fst (nat_snd (nat_snd n)))) with (k := k) as (f & Hf).
    rewrite <- nat_pair_fst_snd.
    rewrite <- (nat_pair_fst_snd (nat_snd n)) at 2.
    rewrite <- (nat_pair_fst_snd (nat_snd (nat_snd n))) at 2.
    rewrite H, <- Heqp.
    generalize (nat_fst (nat_snd (nat_snd n))) (nat_snd (nat_snd (nat_snd n))); intros x y.
    apply le_lt_trans with (<<2;<<x;y>>>>).
    apply le_trans with (2 := nat_pair_le_rt _ _), nat_pair_le_lft.
    apply nat_pair_lt_rt.
    apply nat_pair_le_lft; auto.
    destruct (IHn (nat_snd (nat_snd (nat_snd n)))) with (k := k) as (a & Ha).
    rewrite <- nat_pair_fst_snd.
    rewrite <- (nat_pair_fst_snd (nat_snd n)) at 2.
    rewrite <- (nat_pair_fst_snd (nat_snd (nat_snd n))) at 2.
    rewrite H, <- Heqp.
    generalize (nat_fst (nat_snd (nat_snd n))) (nat_snd (nat_snd (nat_snd n))); intros x y.
    apply le_lt_trans with (<<2;<<x;y>>>>).
    apply le_trans with (2 := nat_pair_le_rt _ _), nat_pair_le_rt.
    apply nat_pair_lt_rt.
    apply nat_pair_le_lft; auto.
    exists (f @ a); intros [ g | g b | v ] H'; subst n.
    simpl in Heqp; rewrite nat_snd_pair_eq, nat_fst_pair_eq in Heqp; discriminate Heqp.
    f_equal. 
    apply Hf; simpl; rewrite nat_snd_pair_eq, nat_snd_pair_eq, nat_fst_pair_eq; auto.
    apply Ha; simpl; rewrite nat_snd_pair_eq, nat_snd_pair_eq, nat_snd_pair_eq; auto.
    simpl in Heqp; rewrite nat_snd_pair_eq, nat_fst_pair_eq in Heqp; discriminate Heqp.
    
    destruct (le_lt_dec k (nat_snd (nat_snd n))) as [ C | C ].
    
    exists (\\ £ pos0); intros [ f | f a | v ]; simpl; intros; subst n.
    rewrite nat_snd_pair_eq, nat_fst_pair_eq in Heqp; discriminate Heqp.
    rewrite nat_snd_pair_eq, nat_fst_pair_eq in Heqp; discriminate Heqp.
    do 2 rewrite nat_snd_pair_eq in C.
    exfalso; generalize (pos2nat_prop v); omega.
    
    exists (£ (nat2pos C)); intros [ f | f a | v ]; simpl; intros; subst n.
    rewrite nat_snd_pair_eq, nat_fst_pair_eq in Heqp; discriminate Heqp.
    rewrite nat_snd_pair_eq, nat_fst_pair_eq in Heqp; discriminate Heqp.
    f_equal.
    revert C; do 2 rewrite nat_snd_pair_eq; intros C.
    apply pos2nat_inj, pos2nat_nat2pos.
  Qed.
  
  Definition uterm_dec n k := proj1_sig (@uterm_dec_rec n k).
  
  Fact uterm_dec_prop k (t : uterm k) : uterm_dec (uterm_enc t) k = t.
  Proof.
    apply (proj2_sig (uterm_dec_rec (uterm_enc t) k) _ eq_refl).
  Qed.
  
  (* The difficulty here will be to show that fun n => uterm_enc (uterm_dec n k)
     is primitive recursive for any k *)

End uterm_enc.

Extraction "uterm_decode.ml" uterm_enc uterm_dec.

 