(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import List Relations Wellfounded.

Require Import utils pos.
Require Import uterm uterm_beta uterm_beta_eq uterm_normal uterm_head.

Set Implicit Arguments.

Definition ut_solvable_at n (f : uterm n) m (u : uterm m) := exists (s1 : uts_t n m) l, ut_subst f s1 *@ l ~b u.

Definition ut_solvable_1 n (f : uterm n) := forall m u, @ut_solvable_at _ f m u.
Definition ut_solvable_2 n (f : uterm n) := forall m, ut_solvable_at f (ut_id m).
Definition ut_solvable_3 n (f : uterm n) := ut_solvable_at (ut_lift f) (£ (pos_fst n)).
Definition ut_solvable_4 n (f : uterm n) := forall m u, exists (v : uterm m) k, ut_subst f (fun _ => v) @* map (fun _ => v) (pos_list k) ~b u.

Notation ut_solvable := ut_solvable_1.

Lemma ut_solvable_1_2 n f : @ut_solvable_1 n f -> ut_solvable_2 f.
Proof. intros H m; apply H. Qed.

Lemma ut_solvable_2_3 n f : @ut_solvable_2 n f -> ut_solvable_3 f.
Proof. 
  intros H; generalize (H n); intros (s1 & l & H').
  apply ut_beta_normal_eq in H'.
  2: apply ut_id_normal.
  apply ut_beta_rt_lift in H'.
  apply (ut_beta_rt_lft (£ (pos_fst _))) in H'.
  assert (ut_lift (ut_subst f s1 *@ l) @ £(pos_fst n) -b>> £(pos_fst _)) as H1.
    constructor 3 with (1 := H').
    constructor 1.
    apply (ut_beta_id (£ (pos_fst n))).
  clear H'.
  rewrite ut_lift_app_rec_rev in H1.
  rewrite ut_lift_subst_eq, ut_subst_comp in H1.
  revert H1.
  match goal with |- (ut_subst _ ?s1 *@ ?s2) @ _ -b>> _ -> _ => set (r1 := s1); set (r2 := s2) end.
  intros H1.
  exists (uts_cons (ut_id _) r1), (£ (pos_fst _)::r2).
  rewrite <- ut_lift_subst.
  apply clos_rt_clos_rst, H1.
Qed.

Lemma ut_solvable_3_1 n f : @ut_solvable_3 n f -> ut_solvable_1 f.
Proof.
  intros (s1 & l & H').
  apply ut_beta_normal_eq in H'.
  2: intro; apply ut_beta_var_inv.
  intros m u.
  set (s3 := (fun _ : pos (S n) => u) : uts_t (S n) m).
  apply ut_beta_rt_subst with (s := s3) in H'.
  rewrite ut_subst_app_rec_rev in H'.
  rewrite ut_subst_comp in H'.
  unfold ut_lift in H'.
  rewrite ut_subst_ren in H'.
  apply clos_rt_clos_rst in H'.
  apply rst_sym in H'.
  exists (fun p => ut_subst (s1 (pos_nxt _ p)) s3).
  exists (map (fun x => ut_subst x s3) l).
  apply ut_beta_eq_sym, H'.
Qed.

Lemma ut_subst_solvable n f m (s : uts_t n m) : ut_solvable (ut_subst f s) -> ut_solvable f.
Proof.
  intros H k u; specialize (H k u).
  destruct H as (s' & l & H).
  rewrite ut_subst_comp in H.
  exists (fun p => ut_subst (s p) s'), l; auto.
Qed.

(* x @ x  solvable but not (x @ x) { x => \x, x@x }*)

Lemma ut_ren_solvable n f m (r : pos_map n m) : ut_solvable (ut_ren f r) -> ut_solvable f.
Proof.
  rewrite ut_ren_subst_eq; apply ut_subst_solvable.
Qed.

Lemma ut_lift_solvable n f : @ut_solvable n f <-> ut_solvable (ut_lift f).
Proof.
  split.
  intros H m u;
  destruct (H _ u) as (s1 & l & H').
  rewrite ut_lift_subst with (k := ut_id _) in H'.
  exists (uts_cons (ut_id _) s1), l; auto.

  apply ut_ren_solvable.
Qed.

Lemma ut_beta_solvable n f g : f -b-> g -> ( @ut_solvable n f <-> @ut_solvable n g ).
Proof.
  intros Hfg; split; intros H m u.

  destruct (H m u) as ( s1 & l & H1 ).
  exists s1, l.
  apply ut_beta_eq_trans with (2 := H1).
  apply ut_beta_eq_sym, ut_beta_rt_eq.
  constructor 1.
  apply ut_beta_app_rec_rev, ut_beta_subst, Hfg.
  
  destruct (H m u) as ( s1 & l & H1 ).
  exists s1, l.
  apply ut_beta_eq_trans with (2 := H1).
  apply ut_beta_rt_eq.
  constructor 1.
  apply ut_beta_app_rec_rev, ut_beta_subst, Hfg.
Qed.

Fact ut_beta_eq_solvable n f g : f ~b g -> (@ut_solvable n f <-> ut_solvable g).
Proof.
  induction 1 as [ f g H | | f g H IH | f g h H1 H2 IH1 IH2 ].
  apply ut_beta_solvable; auto.
  split; auto.
  split; apply IH.
  rewrite H2; auto.
Qed.

Fact ut_beta_rt_solvable n f g : f -b>> g -> (@ut_solvable n f <-> ut_solvable g).
Proof.
  intros H; apply ut_beta_eq_solvable; revert H; apply ut_beta_rt_eq.
Qed.

Section ut_abs_solvable.

  Let ut_abs_solvable_1 n f : ut_solvable f -> @ut_solvable n (\\ f).
  Proof.
    intros H m u.
    destruct (H _ u) as ( s1 & l & H' ).
    revert H'.
    rewrite ut_subst_S.
    match goal with |- (ut_subst f ?s)[?t] *@ _ ~b _ -> _ => set (s3 := s); set (a := t) end.
    intros H'.
    assert (ut_subst (\\ f) (uts_tail s1) @ a *@ l ~b u) as H1.
      constructor 4 with (2 := H').
      constructor 1.
      apply ut_beta_app_rec_rev.
      simpl.
      apply in_ut_beta_redex.
    clear H'.
    rewrite ut_app_rev_rec_eq in H1.
    red. 
    exists (uts_tail s1), (rev (a::rev l)).
    rewrite ut_app_rev_rec_eq.
    constructor 4 with (2 := H1).
    apply ut_eq_beta_eq.
    rewrite rev_involutive; auto.
  Qed.

  Let ut_abs_solvable_2 n f : @ut_solvable n (\\ f) -> ut_solvable f.
  Proof.
    intros H.
    apply ut_solvable_3_1, ut_solvable_2_3.
    apply ut_solvable_1_2 in H.
    red; intros m.
    destruct (H m) as ( s & l & H1 ).
    
    rewrite ut_app_rev_rec_eq in H1.
    revert H1; generalize (rev l); clear l.
    intros [ | x l ] H1; simpl in H1.
    apply ut_beta_normal_eq in H1.
    2: apply ut_id_normal.
    unfold ut_id in H1.
    apply ut_beta_rt_abs_inv in H1.
    destruct H1 as (g & H1 & H2).
    apply ut_abs_inj in H1; subst.
    red.
    exists (uts_cons (ut_id _) s), nil; simpl.
    apply ut_beta_rt_eq.
    apply ut_beta_rt_subst with (s := uts_first (ut_id _)) in H2.
    apply ut_beta_rt_trans with (2 := H2).
    apply ut_eq_beta_rt.
    rewrite ut_subst_comp.
    apply ut_subst_ext.
    intros p; invert pos p; simpl; auto.
    unfold uts_cons, uts_lift; simpl.
    rewrite <- ut_subst_id at 1.
    rewrite ut_lift_subst_eq, ut_subst_comp.
    apply ut_subst_ext; intro; simpl; auto.
    
    exists (uts_cons x s), (rev l).
    apply ut_beta_eq_trans with (2 := H1); clear H1.
    rewrite <- ut_app_rec_rev_eq.
    apply ut_beta_eq_sym.
    apply ut_beta_rt_eq.
    apply ut_beta_rt_app_rec.
    apply ut_beta_redex_rt.
    apply ut_eq_beta_rt.
    unfold ut_reduce.
    rewrite ut_subst_comp.
    apply ut_subst_ext.
    intros p; unfold uts_lift; invert pos p; simpl; auto.
    unfold uts_cons; simpl. 
    rewrite <- ut_subst_id.
    rewrite ut_lift_subst_eq, ut_subst_comp.
    apply ut_subst_ext; intro; simpl; auto.
  Qed.

  Fact ut_abs_solvable n f : @ut_solvable n (\\ f) <-> ut_solvable f.
  Proof. split; auto. Qed.

End ut_abs_solvable.

Lemma ut_app_solvable n f a : @ut_solvable n (f @ a) -> ut_solvable f.
Proof.
  intros H m u.
  destruct (H _ u) as ( s1 & l & H' ).
  simpl in H'.
  rewrite ut_app_rev_rec_eq in H'.
  exists s1, (rev (ut_subst a s1::rev l)).
  constructor 4 with (2 := H').
  apply ut_eq_beta_eq.
  rewrite <- ut_app_rec_rev_eq; auto.
Qed.

Fact ut_var_solvable n (p : pos n) : ut_solvable (£ p).
Proof.
  apply ut_solvable_3_1.
  red; unfold ut_lift; simpl.
  exists (fun _ => £(pos_fst n)), nil; simpl.
  constructor 2.
Qed.

Fact ut_app_rev_solvable n (f : uterm n) l : ut_solvable (f *@ l) -> ut_solvable f.
Proof.
  induction l; simpl; auto.
  intros H; apply ut_app_solvable in H; auto.
Qed.

Fact ut_app_rec_solvable n (f : uterm n) l : ut_solvable (f @* l) -> ut_solvable f.
Proof.
  rewrite ut_app_rec_rev_eq; apply ut_app_rev_solvable.
Qed.

(*

Fact ut_solvable_1_4 n f : @ ut_solvable n f -> ut_solvable_4 f.
Proof.
  induction f as [ n f Hf | n f Hf a Ha | n p Hp ]; intros H.
  apply ut_abs_solvable in H.
  intros m u; specialize (Hf H m u).
  destruct Hf as (v & k & Hf).
  exists v, (S k); simpl.
  rewrite map_map.
  apply ut_beta_eq_trans with (2 := Hf).
  apply ut_beta_rt_eq.
  apply ut_beta_rt_app_rec.
  apply ut_beta_redex_rt.
  apply ut_eq_beta_rt.
  unfold ut_reduce; rewrite ut_subst_comp.
  apply ut_subst_ext.
  intros p; invert pos p; unfold uts_lift; simpl; auto.
  SearchAbout [ ut_subst ut_lift ].
  rewrite <- (ut_subst_id v) at 3.
  rewrite (ut_lift_subst v (@uts_id _) v).
  apply ut_subst_ext.
  intros q; invert pos q; simpl; auto.
  intros m u.
  apply ut_app_solvable in H.
  specialize (Hf H m u).
  destruct Hf as (v & k & Hf).
  

  rewrite ut_lift_subst_eq, ut_subst_comp.
  apply ut_subst_ext.
  intros q; unfold uts_first; simpl.

*)

Section ut_forget.

  Definition ut_forget n m := \\* (ut_id (m+n)).

  Let list_uts n (l : list (uterm n)) : { s : uts_t (length l) n | l = map s (pos_list (length l)) }.
  Proof.
    induction l as [ | x l (s & Hs) ]; simpl.
    exists (fun _ => ut_id _); auto.
    exists (uts_cons x s).
    f_equal; rewrite map_map.
    rewrite Hs at 1.
    apply map_ext.
    intro; auto.
  Defined.
  
  Fact ut_forget_reduce n l : @ut_forget n (length l) *@ l -b>> ut_id _.
  Proof.
    unfold ut_forget.
    destruct (list_uts l) as ( s & Hs ).
    rewrite Hs at 3.
    apply ut_beta_rt_trans with (1 := ut_abs_rec_reduce _ _).
    apply ut_eq_beta_rt; auto.
  Qed.

End ut_forget.

Fact ut_hnf_solvable n f : @ut_hnf n f -> ut_solvable f.
Proof.
  induction f as [ n f IHf | n f Hf g Hg | n p ]; intros H.
  apply ut_hnf_abs in H.
  apply ut_abs_solvable; auto.
  destruct ut_hnf_neutral with (2 := H) as ( p & l & Hp) ; simpl; auto.
  rewrite Hp.
  apply ut_solvable_3_1, ut_solvable_2_3.
  intros m.
  exists (fun _ : pos n => @ut_forget m (length l)), nil.
  rewrite ut_subst_app_rec_rev; simpl.
  apply ut_beta_rt_eq.
  match goal with 
    |- ut_forget _ _ *@ map ?f _ -b>> _ => rewrite <- (map_length f l) at 1
  end.
  apply ut_forget_reduce.
  apply ut_var_solvable.
Qed.

Fact ut_normal_solvable n f : @ut_normal n f -> ut_solvable f.
Proof.
  intros H; apply ut_normal_hnf in H.
  revert H; apply ut_hnf_solvable.
Qed.

Hint Resolve ut_normal_solvable.

Fact ut_normalizable_solvable n f : @ut_normalizable n f -> ut_solvable f.
Proof.
  intros (g & H1 & H2).
  apply ut_beta_eq_solvable with g; auto.
  apply ut_beta_rt_eq; auto.
Qed.

   
